"""Package teneva, module cross_act: compute function of TT-tensors.

This module contains the function "cross_act" which approximates the output of
the given function in the TT-format with input parameters also specified in the
TT-format. Modification of the cross approximation method in the TT-format
(TT-cross) is used.

"""
import numpy as np
import teneva


def cross_act(f, X_list, Y0, e=1.E-6, nswp=10, r=9999, dr=5, dr2=0,
              seed=None, log=False):
    """Compute the output in the TT-format for the function of TT-tensors.

    This is a draft (however, in most cases, the function already works
    successfully. There is a problem in the rank-1 case)!

    This function computes the TT-approximation for the output of the given
    function in the TT-format with input parameters also specified in the
    TT-format. Modification of the cross approximation method in the TT-format
    (TT-cross) is used.

    Args:
        f (function): function f(X) which computes the output elements for the
            given set of input points X, where X is a 2D np.ndarray of the shape
            [samples, D], where D is a number of function inputs. The
            function should return 1D np.ndarray of the length equals to
            samples (i.e., it should be the values of the target function for
            all provided samples).
        X_list (list of lists): several (D) TT-tensors, which are the input
            for the target function (f). Each TT-tensor should be represented
            as a list of its TT-cores. The dimension (d) and mode sizes for
            all tensors must match.
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
            It may be, for example, random TT-tensor, which can be built by the
            "rand" function from teneva: Y0 = teneva.rand(n, r), where n is
            a size of tensor modes (e.g., n = [5, 6, 7, 8, 9] for the
            5-dimensional tensor) and r is related TT-rank (e.g.,
            r = 1). Note that the shape of this tensor should be same as for
            input tensors from X_list.
        e (float): accuracy for SVD truncation and convergence criterion for
            algorithm (> 0). If between iterations the relative rate of
            solution change is less than this value, then the operation of the
            algorithm will be interrupted.
        nswp (int): maximum number of iterations (sweeps) of the algorithm
            (>= 0). One sweep corresponds to a complete pass of all tensor
            TT-cores from right to left and then from left to right.
        r (int): maximum rank for SVD operation (> 0).
        dr (int): rank for AMEN iterations (kickrank; >= 0).
        dr2 (int): additional rank for AMEN iterations (kickrank2; >= 0).
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.
        log (bool): if flag is set, then the information about the progress of
            the algorithm will be printed after each sweep.

    Returns:
        list: TT-Tensor which approximates the output of the function.

    """
    rand = teneva._rand(seed)

    D = len(X_list)               # Number of function inputs
    d = len(X_list[0])            # Dimension of the (any) input tensor
    n = teneva.shape(X_list[0])   # Shape of the (any) input tensor

    # TT-cores of all input tensors (array of shape [d, D] of 3D arrays):
    X = np.array([[G.copy() for G in X] for X in X_list], dtype=object).T

    # TT-tensor for solution:
    Y = teneva.orthogonalize(Y0, d-1)

    # TT-tensor for error:
    Z = teneva.rand(n, dr, seed=rand) if dr > 0 else [None for _ in range(d)]
    Z = teneva.orthogonalize(Z, d-1) if dr > 0 else Z

    # Interface matrices:
    Rx  = _inter_build(d, D)
    Ry  = _inter_build(d)
    Rz  = _inter_build(d)
    Rxz = _inter_build(d, D)
    Ryz = _inter_build(d)

    # Initialization of interface matrices:
    for i in range(d-1):
        Rx[i+1, :], Ry[i+1], Rz[i+1], Rxz[i+1, :], Ryz[i+1] = _inter_update(
            X[i, :], Y[i], Z[i],
            Rx[i, :], Ry[i], Rz[i], Rxz[i, :], Ryz[i], rand, z_rand=True,
            ltr=True)

    i, ltr, swp, e_curr = d, False, 0, 0.
    while swp <= nswp:
        i += 1 if ltr else -1

        G = _func(f, X[i, :], Rx[i, :], Rx[i+1, :])
        G = teneva.core_dot_inv(G, Ry[i], ltr=False)
        G = teneva.core_dot_inv(G, Ry[i+1], ltr=True)
        e_curr = max(e_curr, teneva.accuracy(Y[i], G))
        Y[i], U, V = _svd(G, d, e, r, ltr=ltr)
        U1 = U if ltr else V
        U2 = V if ltr else U

        if not ltr and i == 0:  # We reached the first mode (half of sweep)
            i, ltr = -1, True
            continue
        if ltr and i == d-1:    # We reached the last mode (end of sweep)
            _log(Y, swp, e_curr, log)
            if e_curr < e:
                break
            i, ltr, swp, e_curr = d, False, swp+1, 0,
            continue

        if dr > 0:              # "AMEN-like" operations:
            Gzy = _func(f, X[i, :],
                Rx[i] if ltr else Rxz[i, :],
                Rxz[i+1, :] if ltr else Rx[i+1])
            Gdz = teneva.core_dot(Y[i],
                Ryz[i+1 if ltr else i],
                ltr)
            Gzy = _amen_z(Gzy, Gdz,
                Ry[i if ltr else i+1],
                Rz[i+1 if ltr else i],
                dr, None, True, ltr, rand)

            Gz = _func(f, X[i], Rxz[i], Rxz[i+1])
            Gdz = teneva.core_dot(Gdz,
                Ryz[i if ltr else i+1],
                not ltr)
            Z[i] = _amen_z(Gz, Gdz,
                Rz[i],
                Rz[i+1],
                dr, dr2, False, ltr, rand)

            U1, U2 = _amen(Gzy, U1, U2, ltr)

        # Temporary notation for adjacent indices for the compactness:
        jn = i+1 if ltr else i
        jo = i if ltr else i+1
        ju = i+1 if ltr else i-1

        Y[i] = _matrix_to_core(U1, *Y[i].shape, ltr)

        Y[ju] = teneva.core_dot(Y[ju], U2, not ltr)

        Rx[jn, :], Ry[jn], Rz[jn], Rxz[jn, :], Ryz[jn] = _inter_update(
            X[i, :], Y[i], Z[i],
            Rx[jo, :], Ry[jo], Rz[jo], Rxz[jo, :], Ryz[jo],
            rand, False, ltr)

    return Y


def _amen(G, U1, U2, ltr=True):
    r1, n, r2 = G.shape

    G = teneva._reshape(G, (r1*n, r2) if ltr else (r1, n*r2))
    G = G if ltr else G.T

    U1, U2_add = np.linalg.qr(np.hstack((U1, G)))

    U2 = np.hstack((U2, np.zeros((U2.shape[0], G.shape[1]), dtype=float)))
    U2 = U2_add @ U2.T if ltr else U2 @ U2_add.T

    return U1, U2


def _amen_z(G, dG, R1, R2, dr, dr2=None, is_dz=False, ltr=True, rand=None):
    r1, n, r2 = G.shape

    if not is_dz:
        G = G - dG
    G = teneva.core_dot_inv(G, R1, not ltr if is_dz else False)

    if is_dz:
        G = G - dG
    G = teneva.core_dot_inv(G, R2, ltr if is_dz else True)

    _, U, V = _svd(G, eps=None, rmax=dr, ltr=ltr)
    G = U if ltr else V.T

    G = teneva._reshape(G, (r1, n, G.shape[1]) if ltr else (G.shape[0], n, r2))

    if not is_dz:
        G = teneva.core_qr_rand(G, dr2, ltr, rand)

    return G


def _func(f, G, R1, R2):
    args = []
    for (G_, R1_, R2_) in zip(G, R1, R2):
        Q_ = teneva.core_dot(G_, R1_, ltr=False)
        Q_ = teneva.core_dot(Q_, R2_, ltr=True)
        args.append(teneva._reshape(Q_, -1))
    res = f(np.array(args).T)

    # Note, all R1/R2 have the same ranks, due to usage of "core_dot_maxvol"
    # in "_inter_update" function, hence we know the shape of result:
    n = G[0].shape[1]
    r1 = R1[0].shape[0] if isinstance(R1[0], np.ndarray) else 1
    r2 = R2[0].shape[-1] if isinstance(R2[0], np.ndarray) else 1
    return teneva._reshape(res, (r1, n, r2))


def _inter_build(d, D=None):
    s = d+1 if D is None else (d+1, D)
    R = np.zeros(s, dtype=object)
    R[0] = np.ones((1, 1) if D is None else D, dtype=float)
    R[d] = np.ones((1, 1) if D is None else D, dtype=float)
    return R


def _inter_update(Gx, Gy, Gz, Rx, Ry, Rz, Rxz, Ryz, rand, z_rand=False,
                  ltr=True):
    Ry, ind = teneva.core_dot_maxvol(Gy, Ry, None, not ltr)
    Rx = [teneva.core_dot_maxvol(G, R, ind, not ltr)[0]
        for (G, R) in zip(Gx, Rx)]
    if Gz is not None:
        r1, n, r2 = Gz.shape
        ind = None
        if z_rand:
            perm = rand.permutation(r1*n if ltr else n*r2)
            ind = perm[:(r2 if ltr else r1)]
        Rz, ind = teneva.core_dot_maxvol(Gz, Rz, ind, not ltr)
        Ryz, _ = teneva.core_dot_maxvol(Gy, Ryz, ind, not ltr)
        Rxz = [teneva.core_dot_maxvol(G, R, ind, not ltr)[0]
            for (G, R) in zip(Gx, Rxz)]
    return Rx, Ry, Rz, Rxz, Ryz


def _log(Y, swp, e, log=False):
    if log:
        text = f'== cross-act # {swp+1:-4d} | '
        text += f'e: {e:-8.1e} | r: {teneva.erank(Y):-5.1f}'
        print(text)


def _matrix_to_core(M, r1, n, r2, ltr=True):
    if ltr:
        r2 = M.shape[1]
    else:
        r1 = M.shape[1]
    return teneva._reshape(M if ltr else M.T, (r1, n, r2))


def _rank_trunc(s, eps):
    if eps <= 0.:
        return len(s)
    s = np.cumsum(abs(s[::-1])**2)[::-1]
    r = [i for i in range(len(s)) if s[i] < eps**2]
    return len(s) if len(r) == 0 else np.amin(r)


def _svd(G, d=1, eps=1.E-6, rmax=9999999, is_qr=False, ltr=True):
    r1, n, r2 = G.shape
    G = teneva._reshape(G, (r1*n, r2) if ltr else (r1, n*r2))

    if is_qr: # Note: this is never used in the current code.
        U, V = np.linalg.qr(G if ltr else G.T)
        U = U if ltr else U.T
        V = V.T if ltr else V
        r = U.shape[1]
        s = np.ones(r)
    else:
        U, s, V = np.linalg.svd(G, full_matrices=False)
        V = V.T
        r = _rank_trunc(s, eps/np.sqrt(d) * np.linalg.norm(s)) if eps else rmax
        r = min(r, rmax, len(s))

    s = np.diag(s[:r])
    U = U[:, :r] if ltr else U[:, :r] @ s
    V = V[:, :r] @ s if ltr else V[:, :r]

    G = teneva._reshape(U @ V.T, (r1, n, r2))

    return G, U, V



"""Package teneva, module vectors: various useful QTT-vectors.

This module contains the collection of functions for explicit construction of
various useful QTT-vectors (delta function and others).

"""
import numpy as np
import teneva


def vector_delta(q, i, v=1.):
    """Build QTT-vector that is zero everywhere except for a given index.

    Construct a QTT-vector of the length 2^q with only one nonzero element in
    position i, that is equal to a given value v.

    Args:
        q (int): quantization level. The resulting vector will have the total
            length 2^q.
        i (int): the index for nonzero element (< 2^q). Note that "negative
            index notation" is supported.
        v (float): the value of the vector at index i.

    Returns:
        list: TT-tensor representing the QTT-vector.

    """
    i = teneva._vector_index_prepare(q, i)
    ind = teneva._vector_index_expand(q, i)
    Y = []
    for k in range(q):
        G = np.zeros((1, 2, 1))
        G[0, ind[k], 0] = 1.
        Y.append(G)
    Y[-1][0, ind[-1], 0] = v
    return Y



"""Tests for teneva (in progress).

Run it from the root of the project as "clear && python test/test.py".

"""
import unittest


from test_act_one import *
from test_als import *
from test_cross import *
from test_func import *
from test_grid import *
from test_maxvol import *
from test_tensors import *


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



"""Package teneva, module transformation: transformation of TT-tensors.

This module contains the functions for orthogonalization, truncation and
transformation into full (numpy) format of the TT-tensors.

"""
import numpy as np
import scipy as sp
import teneva


def full(Y):
    """Export TT-tensor to the full (numpy) format.

    Args:
        Y (list): TT-tensor.

    Returns:
        np.ndarray: multidimensional array related to the given TT-tensor.

    Note:
         This function can only be used for relatively small tensors, because
         the resulting tensor will have n^d elements and may not fit in memory
         for large dimensions.

    """
    Z = Y[0].copy()
    for i in range(1, len(Y)):
        Z = np.tensordot(Z, Y[i], 1)

    if Z.shape[0] == 1:
        Z = Z[0, ...]

    if Z.shape[-1] == 1:
        Z = Z[..., 0]

    return Z


def full_matrix(Y):
    """Export QTT-matrix to the full (numpy) format.

    Args:
        Y (list): TT-tensor of dimension q and mode size 4, which
            represents the QTT-matrix of the shape 2^q x 2^q.

    Returns:
        np.ndarray: the matrix of the shape 2^q x 2^q.

    Note:
         This function can only be used for relatively small mode size, because
         the resulting matrix may not fit in memory otherwise.

    """
    q = len(Y)

    Z = full(Y)
    Z = Z.reshape([2, 2]*q, order='F')

    prm = np.hstack((np.arange(0, 2*q, 2), np.arange(1, 2*q, 2)))
    Z = Z.transpose(prm).reshape(2**q, 2**q, order='F')

    return Z


def orthogonalize(Y, k=None, use_stab=False):
    """Orthogonalize TT-tensor.

    Args:
        Y (list): TT-tensor.
        k (int): the leading mode for orthogonalization. The TT-cores 0, 1, ...,
            k-1 will be left-orthogonalized and the TT-cores k+1, k+2, ..., d-1
            will be right-orthogonalized. It will be the last mode by default.
        use_stab (bool): if flag is set, then function will also return the
            second argument p, which is the factor of 2-power.

    Returns:
        list: orthogonalized TT-tensor.

    """
    d = len(Y)
    k = d - 1 if k is None else k

    if k is None or k < 0 or k > d-1:
        raise ValueError('Invalid mode number')

    Z = teneva.copy(Y)
    p = 0

    for i in range(k):
        orthogonalize_left(Z, i, inplace=True)
        if use_stab:
            Z[i+1], p = teneva.core_stab(Z[i+1], p)

    for i in range(d-1, k, -1):
        orthogonalize_right(Z, i, inplace=True)
        if use_stab:
            Z[i-1], p = teneva.core_stab(Z[i-1], p)

    return (Z, p) if use_stab else Z


def orthogonalize_left(Y, i, inplace=False):
    """Left-orthogonalization for TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        i (int): mode for orthogonalization (>= 0 and < d-1).
        inplace (bool): if flag is set, then the original TT-tensor (i.e.,
            the function argument) will be transformed. Otherwise, a copy of
            the TT-tensor will be returned.

    Returns:
        list: TT-tensor with left orthogonalized i-th mode.

    """
    d = len(Y)

    if i is None or i < 0 or i >= d-1:
        raise ValueError('Invalid mode number')

    Z = Y if inplace else teneva.copy(Y)

    r1, n1, r2 = Z[i].shape
    G1 = teneva._reshape(Z[i], (r1 * n1, r2))
    Q, R = np.linalg.qr(G1, mode='reduced')
    Z[i] = teneva._reshape(Q, (r1, n1, Q.shape[1]))

    r2, n2, r3 = Z[i+1].shape
    G2 = teneva._reshape(Z[i+1], (r2, n2 * r3))
    G2 = R @ G2
    Z[i+1] = teneva._reshape(G2, (G2.shape[0], n2, r3))

    return Z


def orthogonalize_right(Y, i, inplace=False):
    """Right-orthogonalization for TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        i (int): mode for orthogonalization (> 0 and <= d-1).
        inplace (bool): if flag is set, then the original TT-tensor (i.e.,
            the function argument) will be transformed. Otherwise, a copy of
            the TT-tensor will be returned.

    Returns:
        list: TT-tensor with right orthogonalized i-th mode.

    """
    d = len(Y)

    if i is None or i <= 0 or i > d-1:
        raise ValueError('Invalid mode number')

    Z = Y if inplace else teneva.copy(Y)

    r2, n2, r3 = Z[i].shape
    G2 = teneva._reshape(Z[i], (r2, n2 * r3))
    R, Q = sp.linalg.rq(G2, mode='economic', check_finite=False)
    Z[i] = teneva._reshape(Q, (Q.shape[0], n2, r3))

    r1, n1, r2 = Z[i-1].shape
    G1 = teneva._reshape(Z[i-1], (r1 * n1, r2))
    G1 = G1 @ R
    Z[i-1] = teneva._reshape(G1, (r1, n1, G1.shape[1]))

    return Z


def truncate(Y, e=1.E-10, r=1.E+12, orth=True, use_stab=False, is_eigh=True):
    """Truncate (round) TT-tensor.

    Args:
        Y (list): TT-tensor wth overestimated ranks.
        e (float): desired approximation accuracy (> 0).
        r (int, float): maximum TT-rank of the result (> 0).
        orth (bool): if the flag is set, then tensor orthogonalization will be
            performed (it is True by default).
        use_stab (bool): if flag is set, then the additional stabilization will
            be used.
        is_eigh (bool): if flag is set, then matrix_svd function will be used
            for truncation of the TT-cores, matrix_skeleton function will be
            used otherwise.

    Returns:
        list: TT-tensor, which is rounded up to a given accuracy e and
        satisfying the rank constraint r.

    """
    d = len(Y)

    if orth:
        if use_stab:
            Z, p = orthogonalize(Y, d-1, True)
            e = e / np.sqrt(d-1) * np.linalg.norm(Z[-1]) # TODO!
        else:
            Z, p = orthogonalize(Y, d-1), 0
            e = e / np.sqrt(d-1) * np.linalg.norm(Z[-1])
    else:
        Z, p = teneva.copy(Y), 0

    for k in range(d-1, 0, -1):
        r1, n, r2 = Z[k].shape
        G = teneva._reshape(Z[k], (r1, n * r2))
        if is_eigh:
            U, V = teneva.matrix_svd(G, e, r)
        else:
            U, V = teneva.matrix_skeleton(G, e, r, rel=False, give_to='r')
        Z[k] = teneva._reshape(V, (-1, n, r2))
        Z[k-1] = np.einsum('ijq,ql', Z[k-1], U, optimize=True)

    if use_stab:
        for k in range(d):
            Z[k] *= 2**(p/d)

    return Z



"""Package teneva, module als_func: construct TT-tensor of coefficients.

This module contains the function "als_func" which computes the
TT-approximation of tensor of coefficients (i.e., the TT-Tucker core tensor) by
TT-ALS algorithm, using given random samples.

"""
import numpy as np
from opt_einsum import contract
import scipy as sp
import teneva
from time import perf_counter as tpc


def als_func(X_trn, y_trn, A0, a=-1., b=+1., nswp=50, e=1.E-16, info={}, *,
             X_vld=None, y_vld=None, e_vld=None, fh=None, lamb=1e-3, n_max=None,
             thr_pow=1.E-6, log=False, update_sol=None):
    """Build TT-Tucker core tensor by TT-ALS from the given samples.

    Args:
        X_trn (np.ndarray): set of train spatial points in the form of array of
            the shape [samples, d], where d is a number of function's input
            dimensions and samples is a size of the train dataset.
        y_trn (np.ndarray): values of the function for inputs X_trn in the
            form of array of the shape [samples].
        A0 (list): TT-tensor, which is the initial approximation for algorithm.
            It should have the equal size for all modes.
        a (float): grid lower bounds for each dimension (should be the same for
            all dimensions in the current version). It is not used if fh
            argument is given.
        b (float): grid upper bounds for each dimension (should be the same for
            all dimensions in the current version). It is not used if fh
            argument is given.
        nswp (int): number of ALS iterations (sweeps). If e or e_vld
            parameter is set, then the real number of sweeps may be less (see
            info dict with the exact number of performed sweeps).
        e (float): optional algorithm convergence criterion (> 0). If between
            iterations (sweeps) the relative rate of solution change is less
            than this value, then the operation of the algorithm will be
            interrupted.
        info (dict): an optionally set dictionary, which will be filled with
            reference information about the process of the algorithm operation.
            At the end of the function work, it will contain parameters: e -
            the final value of the convergence criterion; e_vld - the final
            error on the validation dataset; nswp - the real number of
            performed iterations (sweeps); stop - stop type of the algorithm
            (nswp, e or e_vld).
        X_vld (np.ndarray): optional spatial points for items of validation
            dataset in the form of array of the shape [samples_vld, d], where
            samples_vld is a size of the validation dataset.
        y_vld (np.ndarray): optional values of the function for spatial points
            X_vld of validation dataset in the form of array of the shape
            [samples_vld].
        e_vld (float): optional algorithm convergence criterion (> 0). If
            after sweep, the error on the validation dataset is less than this
            value, then the operation of the algorithm will be interrupted.
        fh (function): optional function, that generates a line in the H
            matrices in the TT-Tucker. If it is not set, then a and b
            arguments should be provided (the Chebyshev interpolation will be
            used in this case).
        lamb (float): regularization parameter for least squares.
        n_max (int): optional maximum mode size for coefficients' tensor. If
            the parameter is set, then a dynamic search for the optimal value
            will be carried out.
        thr_pow (float): optional parameter for dynamic search of the optimal
            value of the mode size for coefficients' tensor.
        log (bool): if flag is set, then the information about the progress of
            the algorithm will be printed after each sweep.

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor
            of interpolation coefficients (i.e., TT-Tucker core tensor).

    """
    _time = tpc()

    if update_sol is not None:
        assert lamb is not None, "Cannot update core without learning rate"

    info.update({'r': teneva.erank(A0), 'e': -1, 'e_vld': -1, 'nswp': 0,
        'stop': None})

    X_trn = np.asanyarray(X_trn, dtype=float)
    y_trn = np.asanyarray(y_trn, dtype=float)

    m = X_trn.shape[0]
    d = X_trn.shape[1]
    n = [G.shape[1] for G in A0]

    Y = teneva.copy(A0)

    if fh is None:
        is_cheb = True
        fh_size = n_max or Y[0].shape[1]
        fh = [lambda X: teneva.func_basis(
            teneva.poi_scale(X, a, b, kind='cheb'), fh_size)]*d
    else:
        try:
            assert len(fh) == d, "Number of functions must be the same as TT-dimension"
        except TypeError:
            fh = [fh]*d

        is_cheb = False

    Yl = [np.ones((m, A0[k].shape[0])) for k in range(d)]
    Yr = [np.ones((A0[k].shape[2], m)) for k in range(d)]

    H = [fhi(x).T for fhi, x in zip(fh, X_trn.T)]
    del X_trn # For test and for memory

    for k, (c, h) in enumerate(zip(Y, H)):
        n_max_curr = h.shape[-1]
        if n_max_curr != c.shape[1]:
            G = np.zeros((c.shape[0], n_max_curr, c.shape[2]))
            G[:, :c.shape[1], :] = c
            Y[k] = G

    if X_vld is not None and y_vld is not None:
        y_our = teneva.func_get(X_vld, Y, funcs=fh)
        y_vld_nrm = np.linalg.norm(y_vld)
        info['e_vld'] = np.linalg.norm(y_our - y_vld) / y_vld_nrm
    teneva._info_appr(info, _time, nswp, e, e_vld, log)

    for k in range(d-1, 0, -1):
        contract('ik,rkq,qi->ri', H[k][:, :n[k]], Y[k][:, :n[k], :], Yr[k],
            out=Yr[k-1])

    while True:
        Yold = teneva.copy(Y)
        nold = list(n)

        for lr in [1, -1]:
            rng = range(0, d-1, +1) if lr == 1 else range(d-1, 0, -1)
            for k in rng:
                n_max_cur = H[k].shape[-1]
                # Temporary increase max pow of poly
                n_k = min(n[k] + 1, n_max_cur)

                if update_sol is not None:
                    Y[k][:, n[k]:, :] = 0

                n[k] = n_k =_optimize_core(Y[k][:, :n_k, :], y_trn,
                    Yl[k], Yr[k], H[k][:, :n_k], n_max_cur, thr_pow, lamb=lamb, update_sol=update_sol)
                Hk = H[k][:, :n_k]

                if lr == 1:
                    contract('jr,jk,krl->jl', Hk, Yl[k], Y[k][:, :n_k, :],
                        out=Yl[k+1])
                else:
                    contract('jr,irk,kj->ij', Hk, Y[k][:, :n_k, :], Yr[k],
                        out=Yr[k-1])

        for nold_k, c in zip(nold, Yold):
            c[:, nold_k:, :] = 0.
        for n_k, c in zip(n, Y):
            c[:, n_k:, :] = 0.

        info['nswp'] += 1
        info['r'] = teneva.erank(Y)
        info['e'] = teneva.accuracy(Y, Yold)
        if X_vld is not None and y_vld is not None:
            y_our = teneva.func_get(X_vld, Y, funcs=fh)
            info['e_vld'] = np.linalg.norm(y_our - y_vld)  / y_vld_nrm

        if teneva._info_appr(info, _time, nswp, e, e_vld, log):
            Y = [(c if n_k == c.shape[1] else c[:, :n_k, :].copy()) for n_k, c in zip(n, Y)]
            return Y


def _optimize_core(Q, y_trn, Yl, Yr, Hk, n_max, thr_pow, lamb=None, update_sol=None):
    A = contract('li,ik,ij->ikjl', Yr, Yl, Hk).reshape(Yl.shape[0], -1)
    if lamb is None:
        sol, residuals, rank, s = sp.linalg.lstsq(A, y_trn,
            overwrite_a=True, overwrite_b=False, lapack_driver='gelsy')
    else:
        if update_sol is not None:
            y_trn = y_trn - A@(Q.reshape(-1))

        AtA = A.T @ A
        Aty = A.T @ y_trn

        sol, residuals, rank, s = sp.linalg.lstsq(AtA + lamb*np.identity(A.shape[1]), Aty,
            overwrite_a=True, overwrite_b=True, lapack_driver='gelsy')

    if update_sol is None:
        Q[...] = sol.reshape(Q.shape)
    else:
        Q += sol.reshape(Q.shape)


    n_k = Q.shape[1]
    if n_max is not None:
        if n_k > 1 and np.abs(Q[:, -1, :]).max() / np.abs(Q).max() < thr_pow:
            n_k = _optimize_core(Q[:, :-1, :], y_trn,
                Yl, Yr, Hk[:, :-1], n_max, thr_pow, lamb=lamb, update_sol=update_sol)

    return n_k



"""Package teneva, module sample: random sampling for/from the TT-tensor.

This module contains functions for sampling from the TT-tensor and for
generation of random multi-indices and points for learning.

"""
import itertools
import numpy as np
import teneva


def sample(Y, m=1, seed=None, unsert=1.E-10):
    """Sample multi-indices according to given probability TT-tensor.

    Args:
        Y (list): TT-tensor, which represents the discrete probability
            distribution.
        m (int, float): number of samples.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.
        unsert (float): noise parameter.

    Returns:
        np.ndarray: generated multi-indices for the tensor in the form
        of array of the shape [m, d], where d is the dimension of the tensor.

    """
    m = int(m)
    d = len(Y)

    rand = teneva._rand(seed)

    phi = [None]*(d+1)
    phi[-1] = np.ones(1)
    for i in range(d-1, 0, -1):
        phi[i] = np.sum(Y[i], axis=1) @ phi[i+1]

    p = Y[0] @ phi[1]
    p = p.flatten()
    p += unsert
    p = np.maximum(p, 0)
    p = p / p.sum()
    ind = rand.choice(Y[0].shape[1], m, p=p)
    phi[0] = Y[0][0, ind, :] # ind here is an array even if m=1

    res = np.zeros((m, d), dtype=int)
    res[:, 0] = ind

    for i, c in enumerate(Y[1:], start=1):
        p = np.einsum('ma,aib,b->mi', phi[i-1], Y[i], phi[i+1])
        p = np.maximum(p, 0)
        ind = [rand.choice(c.shape[1], p=pi/pi.sum()) for pi in p]
        ind = np.array(ind)
        res[:, i] = ind
        phi[i] = np.einsum('il,lij->ij', phi[i-1], c[:, ind])

    return res


def sample_lhs(n, m, seed=None):
    """Generate LHS multi-indices for the tensor of the given shape.

    Args:
        n (list, np.ndarray): tensor size for each dimension (list or
            np.ndarray of int/float of the length d, where d is the
            dimension of the tensor).
        m (int, float): number of samples.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.


    Returns:
        np.ndarray: generated multi-indices for the tensor in the form of array
        of the shape [m, d], where d is the dimension of the tensor.

    """
    n = np.asanyarray(n, dtype=int)
    m = int(m)
    d = len(n)

    rand = teneva._rand(seed)

    I = np.empty((m, d), dtype=int)
    for i, k in enumerate(n):
        I1 = np.repeat(np.arange(k), m // k)
        I2 = rand.choice(k, m-len(I1), replace=False)
        I[:, i] = np.concatenate([I1, I2])
        rand.shuffle(I[:, i])

    return I


def sample_rand(n, m, seed=None):
    """Generate random multi-indices for the tensor of the given shape.

    Args:
        n (list, np.ndarray): tensor size for each dimension (list or
            np.ndarray of int/float of the length d, where d is the
            dimension of the tensor).
        m (int, float): number of samples.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.

    Returns:
        np.ndarray: generated multi-indices for the tensor in the form of array
        of the shape [m, d], where d is the dimension of the tensor.

    """
    n = np.asanyarray(n, dtype=int)
    m = int(m)
    d = len(n)

    rand = teneva._rand(seed)

    I = np.vstack([rand.choice(np.arange(k), m) for k in n]).T

    return I


def sample_rand_poi(a, b, m, seed=None):
    """Generate random multidimensional points from provided limits.

    Args:
        a (list, np.ndarray): lower grid limits (list or np.ndarray of float
            of the length d, where d is the dimension).
        b (list, np.ndarray): upper grid limits (list or np.ndarray of float
            of the length d, where d is the dimension).
        m (int, float): number of samples.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.

    Returns:
        np.ndarray: generated random multidimensional points in the form of
        array of the shape [m, d], where d is the dimension of the problem.

    """
    d = len(a)
    rand = teneva._rand(seed)
    X = [rand.uniform(a[i], b[i], int(m)) for i in range(d)]
    return np.vstack(X).T


def sample_square(Y, m=1, unique=True, seed=None, m_fact=5, max_rep=100,
                  float_cf=None):
    """Sample according to given probability TT-tensor (with squaring it).

    Args:
        Y (list): TT-tensor, which represents the discrete probability
            distribution.
        m (int, float): number of samples.
        unique (bool): if True, then unique multi-indices will be generated.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.
        m_fact (int): scale factor to find enough unique samples.
        max_rep (int): number of restarts to find enough unique samples.
        float_cf (float): special parameter (TODO: check).

    Returns:
        np.ndarray: generated multi-indices for the tensor in the form
        of array of the shape [m, d], where d is the dimension of the tensor.

    """
    m = int(m)
    d = len(Y)

    rand = teneva._rand(seed)

    err_msg = 'Can not generate the required number of samples'

    Z, p = teneva.orthogonalize(Y, 0, use_stab=True)

    G = Z[0]
    r1, n, r2 = G.shape

    if float_cf is not None:
        n, nold = n * float_cf, n
        G = _extend_core(G, n)

    Q = G.reshape(n, r2)

    m1 = m_fact*m if unique else m
    Q, I1 = _sample_core_first(Q, teneva._range(n), m1, rand)
    I = np.empty([I1.shape[0], d])
    I[:, 0] = I1[:, 0]

    for di, G in enumerate(Z[1:], start=1):
        r1, n, r2 = G.shape

        if float_cf is not None:
            n, nold = n * float_cf, n
            G = _extend_core(G, n)

        Qtens = np.einsum('kr,riq->kiq', Q, G, optimize='optimal')
        Q = np.empty([Q.shape[0], r2])

        for im, qm, qnew in zip(I, Qtens, Q):
            norms = np.sum(qm**2, axis=1)
            norms /= norms.sum()

            i_cur = im[di] = rand.choice(n, size=1, p=norms)
            qnew[:] = qm[i_cur]

    if unique:
        I = np.unique(I, axis=0)
        if I.shape[0] < m:
            if max_rep < 0 or m_fact > 1000000:
                raise ValueError(err_msg)
            return sample_square(Y, m, True, seed, 2*m_fact, max_rep-1,
                float_cf=float_cf)
        else:
            np.random.shuffle(I)

    I = I[:m]

    if I.shape[0] != m:
        raise ValueError(err_msg)

    if float_cf is not None:
        I = I / float_cf
    else:
        I = I.astype(int)

    return I


def sample_tt(n, r=4, seed=None):
    """Generate special samples for the tensor of the shape n.

    Generate special samples (multi-indices) for the tensor, which are the best
    (in many cases) for the subsequent construction of the TT-tensor.

    Args:
        n (list, np.ndarray): tensor size for each dimension (list or
            np.ndarray of int/float of the length d).
        r (int): expected TT-rank of the tensor. The number of generated
            samples will be selected according to this value.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.

    Returns:
        (np.ndarray, np.ndarray, np.ndarray): generated multi-indices for the
        tensor in the form of array of the shape [samples, d], starting
        poisitions in generated samples for the corresponding dimensions in the
        form of array of the shape [d+1] and numbers of points for the right
        unfoldings in generated samples in the form of array of the shape [d].

    Note:
        The resulting number of samples will be chosen adaptively based on the
        specified expected TT-rank (r).

    """
    def one_mode(sh1, sh2, rng):
        res = []
        if len(sh2) == 0:
            lhs_1 = sample_lhs(sh1, r, seed)
            for n in range(rng):
                for i in lhs_1:
                    res.append(np.concatenate([i, [n]]))
            len_1, len_2 = len(lhs_1), 1
        elif len(sh1) == 0:
            lhs_2 = sample_lhs(sh2, r, seed)
            for n in range(rng):
                for j in lhs_2:
                    res.append(np.concatenate([[n], j]))
            len_1, len_2 = 1, len(lhs_2)
        else:
            lhs_1 = sample_lhs(sh1, r, seed)
            lhs_2 = sample_lhs(sh2, r, seed)
            for n in range(rng):
                for i, j in itertools.product(lhs_1, lhs_2):
                    res.append(np.concatenate([i, [n], j]))
            len_1, len_2 = len(lhs_1), len(lhs_2)
        return res, len_1, len_2

    I, idx, idx_many = [], [0], []
    for i in range(len(n)):
        pnts, len_1, len_2 = one_mode(n[:i], n[i+1:], n[i])
        I.append(pnts)
        idx.append(idx[-1] + len(pnts))
        idx_many.append(len_2)

    return np.vstack(I), np.array(idx), np.array(idx_many)


def _extend_core(G, n):
    r1, nold, r2 = G.shape
    Gn = np.empty([r1, n, r2])

    for i1 in range(r1):
        for i2 in range(r2):
            Gn[i1, :, i2] = np.interp(np.arange(n)*(nold - 1)/(n - 1),
                range(nold), G[i1, :, i2])
    return Gn


def _sample_core_first(Q, I, m, rand):
    n = Q.shape[0]

    norms = np.sum(Q**2, axis=1)
    norms /= norms.sum()

    ind = rand.choice(n, size=m, p=norms, replace=True)

    return Q[ind, :], I[ind, :]



"""Package teneva, module data: functions for working with datasets.

This module contains functions for working with datasets, including
"accuracy_on_data" function.

"""
import numpy as np
import teneva


def accuracy_on_data(Y, I_data, y_data, e_trunc=None):
    """Compute the relative error of TT-tensor on the dataset.

    Args:
        I_data (np.ndarray): multi-indices for items of dataset in the form of
            array of the shape [samples, d].
        y_data (np.ndarray): values for items related to I_data of dataset in
            the form of array of the shape [samples].
        e_trunc (float): optional truncation accuracy. If this parameter is
            set, then sampling will be performed from the rounded TT-tensor.

    Returns:
        float: the relative error.

    Note:
        If I_data or y_data is not provided, the function will return -1.

    """
    if I_data is None or y_data is None:
        return -1.

    I_data = np.asanyarray(I_data, dtype=int)
    y_data = np.asanyarray(y_data, dtype=float)

    if e_trunc is not None:
        Y = teneva.truncate(Y, e_trunc)

    y = teneva.get_many(Y, I_data)
    return np.linalg.norm(y - y_data) / np.linalg.norm(y_data)


def cache_to_data(cache={}):
    """Transform cache of the TT-cross into I, Y data arrays.

    Args:
        cache (dict): cache of the TT-cross (see "cross" function), that
            contains the requested function values and related tensor
            multi-indices.

    Returns:
        (np.ndarray, np.ndarray): tensor multi-indices (I_data; in the form of
        array of the shape [samples, dimension]) and related function values
        (y_data; in the form of array of the shape [samples]).

    """
    I_data = np.array([i for i in cache.keys()], dtype=int)
    y_data = np.array([y for y in cache.values()])
    return I_data, y_data



"""Package teneva, module act_two: operations with a pair of TT-tensors.

This module contains the basic operations with a pair of TT-tensors "(Y1, Y2)",
including "add", "mul", "sub", etc.

"""
import numpy as np
import teneva


def accuracy(Y1, Y2):
    """Compute || Y1 - Y2 || / || Y2 || for tensors in the TT-format.

    Args:
        Y1 (list): TT-tensor.
        Y2 (list): TT-tensor.

    Returns:
        float: the relative difference between two tensors.

    Note:
        Also, for convenience, tensors in numpy format can be passed, in this
        case the norm will be calculated using the function "linalg.norm".

    """
    if isinstance(Y1, np.ndarray):
        return np.linalg.norm(Y1 - Y2) / np.linalg.norm(Y2)

    z1, p1 = teneva.norm(sub(Y1, Y2), use_stab=True)
    z2, p2 = teneva.norm(Y2, use_stab=True)

    if p1 - p2 > 500:
        return 1.E+299
    if p1 - p2 < -500:
        return 0.

    c = 2.**(p1 - p2)

    if np.isinf(c) or np.isinf(z1) or np.isinf(z2) or abs(z2) < 1.E-100:
        return -1 # TODO: check

    return c * z1 / z2


def add(Y1, Y2):
    """Compute Y1 + Y2 in the TT-format.

    Args:
        Y1 (int, float, list): TT-tensor (or it may be int/float).
        Y2 (int, float, list): TT-tensor (or it may be int/float).

    Returns:
        list: TT-tensor, which represents the element wise sum of Y1 and Y2.
        If both Y1 and Y2 are numbers, then result will be int/float number.

    """
    if teneva._is_num(Y1) and teneva._is_num(Y2):
        return Y1 + Y2
    elif teneva._is_num(Y1):
        Y1 = teneva.const(teneva.shape(Y2), Y1)
    elif teneva._is_num(Y2):
        Y2 = teneva.const(teneva.shape(Y1), Y2)

    n, r1, r2, Y = teneva.shape(Y1), teneva.ranks(Y1), teneva.ranks(Y2), []
    for i, (G1, G2, k) in enumerate(zip(Y1, Y2, n)):
        if i == 0:
            G = np.concatenate([G1, G2], axis=2)
        elif i == len(n) - 1:
            G = np.concatenate([G1, G2], axis=0)
        else:
            r1_l, r1_r = r1[i:i+2]
            r2_l, r2_r = r2[i:i+2]
            Z1 = np.zeros([r1_l, k, r2_r])
            Z2 = np.zeros([r2_l, k, r1_r])
            L1 = np.concatenate([G1, Z1], axis=2)
            L2 = np.concatenate([Z2, G2], axis=2)
            G = np.concatenate([L1, L2], axis=0)
        Y.append(G)

    return Y


def mul(Y1, Y2):
    """Compute element wise product Y1 * Y2 in the TT-format.

    Args:
        Y1 (int, float, list): TT-tensor (or it may be int/float).
        Y2 (int, float, list): TT-tensor (or it may be int/float).

    Returns:
        list: TT-tensor, which represents the element wise product of Y1 and
        Y2. If both Y1 and Y2 are numbers, then result will be float
        number.

    """
    if teneva._is_num(Y1) and teneva._is_num(Y2):
        return Y1 * Y2

    if teneva._is_num(Y1):
        Y = teneva.copy(Y2)
        Y[0] *= Y1
        return Y

    if teneva._is_num(Y2):
        Y = teneva.copy(Y1)
        Y[0] *= Y2
        return Y

    Y = []
    for G1, G2 in zip(Y1, Y2):
        G = G1[:, None, :, :, None] * G2[None, :, :, None, :]
        G = G.reshape([G1.shape[0]*G2.shape[0], -1, G1.shape[-1]*G2.shape[-1]])
        Y.append(G)

    return Y


def mul_scalar(Y1, Y2, use_stab=False):
    """Compute scalar product for Y1 and Y2 in the TT-format.

    Args:
        Y1 (list): TT-tensor.
        Y2 (list): TT-tensor.
        use_stab (bool): if flag is set, then function will also return the
            second argument p, which is the factor of 2-power.

    Returns:
        float: the scalar product.

    """
    v = None
    p = 0

    for i, (G1, G2) in enumerate(zip(Y1, Y2)):
        G = G1[:, None, :, :, None] * G2[None, :, :, None, :]
        G = G.reshape([G1.shape[0]*G2.shape[0], -1, G1.shape[-1]*G2.shape[-1]])
        G = np.sum(G, axis=1)
        v = G.copy() if i == 0 else v @ G

        if use_stab:
            v, p = teneva.core_stab(v, p)

    v = v.item()

    return (v, p) if use_stab else v


def outer(Y1, Y2):
    """Compute outer product of two TT-tensors.

    Args:
        Y1 (list): TT-tensor.
        Y2 (list): TT-tensor.

    Returns:
        list: TT-tensor, which is the outer product of Y1 and Y2.

    Note:
        See also "outer_many" function, which computes outer product of many
        given TT-tensors.

    """
    Y = teneva.copy(Y1)
    Y.extend(teneva.copy(Y2))
    return Y


def sub(Y1, Y2):
    """Compute Y1 - Y2 in the TT-format.

    Args:
        Y1 (int, float, list): TT-tensor (or it may be int/float).
        Y2 (int, float, list): TT-tensor (or it may be int/float).

    Returns:
        list: TT-tensor, which represents the result of the operation Y1-Y2.
        If both Y1 and Y2 are numbers, then result will be float number.

    """
    if teneva._is_num(Y1) and teneva._is_num(Y2):
        return Y1 - Y2

    if teneva._is_num(Y2):
        Y2 = teneva.const(teneva.shape(Y1), -1.*Y2)
    else:
        Y2 = teneva.copy(Y2)
        Y2[0] *= -1.

    return add(Y1, Y2)



"""Package teneva, module vis: visualization methods for tensors.

This module contains the functions for visualization of TT-tensors.

"""
import numpy as np
import teneva


def show(Y):
    """Check and display mode sizes and TT-ranks of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    """
    if not isinstance(Y, list) or len(Y) == 0:
        raise ValueError('Invalid TT-tensor')

    d = len(Y)
    n = []
    r = [1]

    for G in Y:
        if not isinstance(G, np.ndarray) or len(G.shape) != 3:
            raise ValueError('Invalid core for TT-tensor')

        if G.shape[0] != r[-1]:
            raise ValueError('Invalid shape of core for TT-tensor')

        n.append(G.shape[1])
        r.append(G.shape[2])

    if r[-1] != 1:
        raise ValueError('Invalid shape of core for TT-tensor')

    text1 = f'TT-tensor {d:-5d}D : '
    text2 = f'<rank>  = {teneva.erank(Y):-6.1f} : '

    for k in range(d):
        text1 += ' ' * max(0, len(text2)-len(text1)-1)
        text1 += f'|{n[k]}|'

        if k < d-1:
            text2 += ' ' * (len(text1)-len(text2)-1)
            text2 += f'\\{r[k+1]}/'

    print(text1 + '\n' + text2)



"""Package teneva, module als: construct TT-tensor, using TT-ALS.

This module contains the function "als" which computes the TT-approximation for
the tensor by TT-ALS algorithm, using given random samples (i.e., the set of
random tensor multi-indices and related tensor values).

"""
import numpy as np
from opt_einsum import contract
import scipy as sp
import teneva
from time import perf_counter as tpc


def als(I_trn, y_trn, Y0, nswp=50, e=1.E-16, info={}, *, I_vld=None, y_vld=None,
        e_vld=None, r=None, r_add=10000, e_adap=1.E-3, lamb=0.001, w=None,
        cb=None, swap_tol=3, allow_swap=False, allow_skip_cores=False,
        use_stab=False, log=False, update_sol=None):
    """Build TT-tensor by TT-ALS method using given random tensor samples.

    Args:
        I_trn (np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d], where d is a number of tensor's
            dimensions and samples is a size of the train dataset.
        y_trn (np.ndarray): values of the tensor for multi-indices I_trn in
            the form of array of the shape [samples].
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        nswp (int): number of ALS iterations (sweeps). If e or e_vld
            parameter is set, then the real number of sweeps may be less (see
            info dict with the exact number of performed sweeps).
        e (float): optional algorithm convergence criterion (> 0). If between
            iterations (sweeps) the relative rate of solution change is less
            than this value, then the operation of the algorithm will be
            interrupted.
        info (dict): an optionally set dictionary, which will be filled with
            reference information about the process of the algorithm operation.
            At the end of the function work, it will contain parameters: e -
            the final value of the convergence criterion; e_vld - the final
            error on the validation dataset; nswp - the real number of
            performed iterations (sweeps); stop - stop type of the algorithm
            (nswp, e, e_vld or cb).
        I_vld (np.ndarray): optional multi-indices for items of validation
            dataset in the form of array of the shape [samples_vld, d], where
            samples_vld is a size of the validation dataset.
        y_vld (np.ndarray): optional values of the tensor for multi-indices
            I_vld of validation dataset in the form of array of the shape
            [samples_vld].
        e_vld (float): optional algorithm convergence criterion. If after
            sweep, the error on the validation dataset is less than this value,
            then the operation of the algorithm will be interrupted.
        r (int): maximum TT-rank for rank-adaptive ALS algorithm. If is None,
            then the TT-ALS with constant rank will be used (in the case of the
            constant rank, its value will be the same as the rank of the
            initial approximation Y0).
        r_add (int): maximum rank grow on one iteration for the rank-adaptive
            ALS algorithm. It is used only if r argument is not None.
        e_adap (float): convergence criterion for rank-adaptive TT-ALS
            algorithm (> 0). It is used only if r argument is not None.
        lamb (float): regularization parameter for least squares.
        w (np.ndarray): optional vector for weights of the input data (it
            should have a length equal to the number of elements in the data
            set). If this vector is used, then lamb parameter should be None.
        cb (function): optional callback function. It will be called after each
            sweep and the accuracy check with the arguments: Y, info and opts,
            where Y is the current approximation (TT-tensor), info is the info
            dictionary and the dictionary opts contains fields Yl, Yr and Yold.
            If the callback returns a true value, then the algorithm will be
            stopped (in the info dictionary, in this case, the stop type of the
            algorithm will be cb).
        swap_tol (int): experimental option.
        allow_swap (bool): experimental flag.
        allow_skip_cores (bool): if there is no data to learn all slices of the,
            TT-cores still work, keeping these slices. If the flag is not
            enabled, then in case of insufficient size of the training dataset,
            an error will be generated.
        use_stab (bool): if the flag is set, then the rank-adaptive method will
            use additional stabilization of the cores.
        log (bool): if flag is set, then the information about the progress of
            the algorithm will be printed after each sweep (and before the
            first sweep).

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor.

    """
    _time = tpc()

    assert r is None or update_sol is None, "Cannot update core of non-constant rank"

    info.update({'e': -1, 'e_vld': -1, 'nswp': 0, 'stop': None})
    info['r'] = teneva.erank(Y0)

    I_trn = np.asanyarray(I_trn, dtype=int)
    y_trn = np.asanyarray(y_trn, dtype=float)

    m = I_trn.shape[0]
    d = I_trn.shape[1]

    if allow_swap:
        msg = 'The option "allow_swap" works only with adaptive rank'
        assert r is not None, msg
        I_trn = I_trn.copy()
        rearrange = np.arange(d)
        info['rearrange'] = rearrange
        print('!!! Note that "allow_swap" is a VERY experimental option')

    Y = teneva.copy(Y0)
    if r is not None:
        Y = teneva.orthogonalize(Y, 0, use_stab)

    if not allow_skip_cores:
        for k in range(d):
            if np.unique(I_trn[:, k]).size != Y[k].shape[1]:
                msg = 'One groundtruth sample is needed for every slice'
                raise ValueError(msg)

    info['e_vld'] = teneva.accuracy_on_data(Y, I_vld, y_vld)
    teneva._info_appr(info, _time, nswp, e, e_vld, log)

    Yl = [np.ones((m, Y[k].shape[0])) for k in range(d)]
    Yr = [np.ones((Y[k].shape[2], m)) for k in range(d)]

    for k in range(d-1, 0, -1):
        i = I_trn[:, k]
        Q = Y[k][:, i, :]
        contract('riq,qi->ri', Q, Yr[k], out=Yr[k-1])

    while True:
        Yold = teneva.copy(Y)
        was_swap = False

        idx_cache = dict()
        for k in range(0, d-1 if r is None else d-2, +1):
            i = I_trn[:, k]

            if r is None:
                Y[k] = _optimize_core(Y[k], i, y_trn, Yl[k], Yr[k],
                    lamb=lamb, w=w, update_sol=update_sol)
                contract('jk,kjl->jl', Yl[k], Y[k][:, i, :], out=Yl[k+1])
            else:
                swaped = {} if allow_swap else None
                r_max = min(r, Y[k].shape[-1] + r_add)
                Y[k], Y[k+1] = _optimize_core_adaptive(Y[k], Y[k+1],
                    i, I_trn[:, k+1], y_trn, Yl[k], Yr[k+1],
                    e_adap, r_max, lamb, w, ltr=True,
                    allow_swap=swaped, swap_tol=swap_tol, cache=idx_cache)
                idx_cache = dict(i1=idx_cache['i2'])
                if allow_swap and swaped.get('swapped', False):
                    print(f'DEBUG | idxs: {k} <-> {k+1}')
                    was_swap = True
                    I_trn[:, [k, k+1]] = I_trn[:, [k+1, k]]
                    i = I_trn[:, k]
                    swap_two = np.arange(len(rearrange))
                    swap_two[k], swap_two[k+1] = swap_two[k+1], swap_two[k]
                    rearrange[:] = swap_two[rearrange]
                    idx_cache = {}
                Yl[k+1] = contract('jk,kjl->jl', Yl[k], Y[k][:, i, :])

        idx_cache = dict()
        for k in range(d-1, 0 if r is None else 1, -1):
            i = I_trn[:, k]

            if r is None:
                Y[k] = _optimize_core(Y[k], i, y_trn, Yl[k], Yr[k],
                    lamb=lamb, w=w, update_sol=update_sol)
                contract('ijk,kj->ij', Y[k][:, i, :], Yr[k], out=Yr[k-1])
            else:
                swaped = {} if allow_swap else None
                r_max = min(r, Y[k-1].shape[-1] + r_add)
                Y[k-1], Y[k] = _optimize_core_adaptive(Y[k-1], Y[k],
                    I_trn[:, k-1], i, y_trn, Yl[k-1], Yr[k],
                    e_adap, r_max, lamb, w, ltr=False,
                    allow_swap=swaped, swap_tol=swap_tol, cache=idx_cache)
                idx_cache = dict(i2=idx_cache['i1'])
                if allow_swap and swaped.get('swapped', False):
                    print(f'DEBUG | idxs: {k} <-> {k-1}')
                    was_swap = True
                    I_trn[:, [k, k-1]] = I_trn[:, [k-1, k]]
                    i = I_trn[:, k]
                    swap_two = np.arange(len(rearrange))
                    swap_two[k], swap_two[k-1] = swap_two[k-1], swap_two[k]
                    rearrange[:] = swap_two[rearrange]
                    idx_cache = {}
                Yr[k-1] = contract('ijk,kj->ij', Y[k][:, i, :], Yr[k])

        info['nswp'] += 1
        info['r'] = teneva.erank(Y)
        info['e'] = 1.E+10 if was_swap else teneva.accuracy(Y, Yold)
        info['e_vld'] = teneva.accuracy_on_data(
            Y, I_vld[:, rearrange] if allow_swap else I_vld, y_vld)

        if cb:
            opts = {'Yold': Yold, 'Yl': Yl, 'Yr': Yr}
            if cb(Y, info, opts) is True:
                info['stop'] = info['stop'] or 'cb'

        if teneva._info_appr(info, _time, nswp, e, e_vld, log):
            return Y


def _lstsq(A, y, lamb=1e-2, w=None, *, overwrite_a=True, update_sol=None):
    if update_sol is not None:
        y = y - A@update_sol

    if lamb is not None:
        if w is not None:
            AW = w[:, None] * A
            AtA = A.T @ AW
            Aty = AW.T @ y
        else:
            AtA = A.T @ A
            Aty = A.T @ y

        return sp.linalg.lstsq(AtA + lamb * np.identity(A.shape[1]), Aty,
                overwrite_a=True, overwrite_b=True, lapack_driver='gelsy')

    else:
        if w is not None:
            A = w[:, None] * A
            y = y * w
        else:
            if not overwrite_a:
                A = np.copy(A)
        return sp.linalg.lstsq(A, y,
            overwrite_a=True, overwrite_b=True, lapack_driver='gelsy')



def _optimize_core(Q, i, y_trn, Yl, Yr, lamb, w, update_sol=None):
    Q = Q.copy()

    for k in range(Q.shape[1]):
        idx = np.where(i == k)[0]
        if not idx.any():
            continue

        lhs = Yr[:, idx].T[:, np.newaxis, :]
        rhs = Yl[idx, :][:, :, np.newaxis]
        A = (lhs * rhs).reshape(len(idx), -1)
        b = y_trn[idx]

        if update_sol is None:
            sol, residuals, rank, s = _lstsq(A, b, lamb=lamb,
                w=w[idx] if w is not None else None, update_sol=None)
            Q[:, k, :] = sol.reshape(Q[:, k, :].shape)
        else:
            sol, residuals, rank, s = _lstsq(A, b, lamb=lamb,
                w=w[idx] if w is not None else None, update_sol=Q[:, k, :].reshape(-1))
            Q[:, k, :] += sol.reshape(Q[:, k, :].shape)


        if False and rank < A.shape[1]: # TODO: check
            print(f'ALS WRN | Bad cond in LSTSQ: {rank} < {A.shape[1]}')

    return Q


def _optimize_core_adaptive(Q1, Q2, i1, i2, y_trn, Yl, Yr, e, r, lamb, w,
                            ltr=True, allow_swap=None, swap_tol=3, cache=None):
    shape = Q1.shape[0], Q2.shape[2]
    shapeQ1 = Q1.shape[:2]
    shapeQ2 = Q2.shape[1:]

    Q = np.empty((Q1.shape[0], Q1.shape[1], Q2.shape[1], Q2.shape[2]))

    cache = {} if cache is None else cache

    try:
        i1_cache = cache['i1']
    except KeyError:
        cache['i1'] = i1_cache = dict()
        for k1 in range(Q1.shape[1]):
            i1_cache[k1] = i1 == k1

    try:
        i2_cache = cache['i2']
    except KeyError:
        cache['i2'] = i2_cache = dict()
        for k2 in range(Q2.shape[1]):
            i2_cache[k2] = i2 == k2

    for k1 in range(Q1.shape[1]):
        for k2 in range(Q2.shape[1]):
            idx = cache['i1'][k1] & cache['i2'][k2]
            if not idx.any():
                continue

            lhs = Yr[:, idx].T[:, np.newaxis, :]
            rhs = Yl[idx, :][:, :, np.newaxis]
            A = (lhs * rhs).reshape(idx.sum(), -1)
            b = y_trn[idx]

            sol, residuals, rank, s = _lstsq(A, b, lamb=lamb,
                w=w[idx] if w is not None else None)
            Q[:, k1, k2, :] = sol.reshape(shape)

            if False and rank < A.shape[1]: # TODO: check
                print(f'ALS WRN | Bad cond in LSTSQ: {rank} < {A.shape[1]}')

    Qs = Q.reshape(np.prod(Q.shape[:2]), -1)
    V1, V2 = teneva.matrix_skeleton(Qs, e, r,
        rel=True, give_to='r' if ltr else 'l')

    rank1 = V1.shape[-1]

    if allow_swap is not None:
        Q = np.transpose(Q, [0, 2, 1, 3])
        Qsr = Q.reshape(np.prod(Q.shape[:2]), -1)
        V1r, V2r = teneva.matrix_skeleton(Qsr, e, r,
            rel=True, give_to='r' if ltr else 'l')
        rank2 = V1r.shape[-1]

        qual1 = _quality_of_decomp(Qs, V1, V2)
        qual2 = _quality_of_decomp(Qsr, V1r, V2r) * swap_tol
        if rank2 < rank1 or qual1 > qual2:
            print(f'DEBUG | ranks: {rank2} < {rank1}, swapping', end=' ')
            allow_swap['swapped'] = True
            V1 = V1r
            V2 = V2r
            shapeQ1 = (Q1.shape[0], Q2.shape[1])
            shapeQ2 = (Q1.shape[1], Q2.shape[-1])

    return V1.reshape(*shapeQ1, -1), V2.reshape(-1, *shapeQ2)


def _quality_of_decomp(Q, V1, V2):
    return np.linalg.norm(V1 @ V2 - Q) / np.linalg.norm(Q)



"""Package teneva, module func: functional TT-format with interpolation.

This module contains the functions for construction of the functional
TT-representation, including Chebyshev interpolation in the TT-format, as well
as calculating the values of the function using the constructed interpolation
coefficients.

"""
from functools import reduce
import numpy as np
import scipy as sp
from scipy.fftpack import dct
from scipy.fftpack import dst
import teneva


def func_basis(X, m=10, kind='cheb'):
    """Compute the basis functions in the given points.

    The function computes values of the first m basis functions (Chebyshev
    polynomials in the current version) in the given points X.

    Args:
        X (np.ndarray): spatial points of interest (it is 2D array of the shape
            [samples, d], where d is a number of dimensions). All points
            should be from the interval [-1, 1] for "cheb" kind and from the
            interval [0, pi] for "sin" kind.
        m (int): maximum order of the basis function (>= 1). The first m
            functions (of the order 0, 1, ..., m-1) will be computed.
        kind (str): kind of the basis ("cheb").

    Returns:
        np.ndarray: values of the basis functions of the order 0, 1, ..., m-1
        in X points (it is 3D array of the shape [m, samples, d]).

    """
    if kind != 'cheb':
        raise NotImplementedError(f'The kind "{kind}" is not supported')

    T = np.ones([m] + list(X.shape))

    if m < 2:
        return T

    T[1, ] = X.copy()
    for k in range(2, m):
        T[k, ] = 2. * X * T[k - 1, ] - T[k - 2, ]

    return T


def func_diff_matrix(a, b, n, m=1, kind='cheb'):
    """Construct the differential matrix (Chebyshev or Sin) of any order.

    The function returns the matrix D (if m=1), which, for the known vector
    y of values of a one-dimensional function on the related grid ("Chebyshev"
    grid if kind is "cheb" or uniform grid if kind is "sin"), gives its first
    derivative, i.e., y' = D y. If the argument m is greater than 1, then the
    function returns a list of matrices corresponding to the first derivative,
    the second derivative, and so on. Note that the derivative error can be
    significant near the boundary of the region.

    Args:
        a (float): grid lower bound.
        b (float): grid upper bound.
        n (int): grid size.
        m (int): the maximum order of derivative.
        kind (str): kind of the basis ("cheb" or "sin").

    Returns:
        list of np.ndarray or np.ndarray: the differential matrices of order 1,
        2, ..., m if m > 1, or only one matrix corresponding to the first
        derivative if m = 1.

    """
    n = int(n)
    k = np.arange(n)

    if kind == 'sin':
        D = np.diag((b - a) / np.pi * np.arange(1, n+1))
        D_list = [D]

        for i in range(m):
            D_list.append(D_list[-1] * D)

    elif kind == 'cheb':
        n1 = int(np.floor(n / 2))
        n2 = int(np.ceil(n / 2))
        th = k * np.pi / (n - 1)

        T = np.tile(th/2, (n, 1))
        DX = 2. * np.sin(T.T + T) * np.sin(T.T - T)
        DX[n1:, :] = -np.flipud(np.fliplr(DX[0:n2, :]))
        DX[range(n), range(n)] = 1.
        DX = DX.T

        Z = 1. / DX
        Z[range(n), range(n)] = 0.

        C = sp.linalg.toeplitz((-1.)**k)
        C[+0, :] *= 2.
        C[-1, :] *= 2.
        C[:, +0] *= 0.5
        C[:, -1] *= 0.5

        D_list = []
        D = np.eye(n)
        for i in range(m):
            D = (i+1) * Z * (C * np.tile(np.diag(D), (n, 1)).T - D)
            D[range(n), range(n)] = -np.sum(D, axis=1)
            l = (2. / (b - a))**(i+1)
            D_list.append(D * l)

    else:
        raise ValueError('Invalid "kind"')

    return D_list[0] if m == 1 else D_list


def func_diff_matrix_apply(A, D, kind='cheb'):
    """Draft. TODO"""
    if kind == 'sin':
        d = np.diag(D)
        return [d[:G.shape[1]] @ G for G in A]

    elif kind == 'cheb':
        raise NotImplementedError()

    else:
        raise ValueError('Invalid "kind"')


def func_get(X, A, a=None, b=None, z=0., funcs=None, kind='cheb',
             skip_out=None):
    """Compute the functional TT-approximation in given points (approx. f(X)).

    Args:
        X (np.ndarray): spatial points of interest (it is 2D array of the shape
            [samples, d], where d is the number of dimensions). In the case of
            only one point, it may be also 1D array.
        A (list): TT-tensor of the interpolation coefficients (it has d
            dimensions).
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.
        z (float): the value for points, which are outside the spatial grid.
        funcs (list of callables): optional custom basis functions.
        kind (str): kind of the basis ("cheb"). It is not used, kept for down
            compatibility.
        skip_out (bool): if flag is set, then the values outside the spatial
            grid will be set to z values.

    Returns:
        np.ndarray: approximated function values in X points (it is 1D array of
        the shape [samples]). In the case of only one input point X, it will be
        float value.

    """
    X = np.asanyarray(X, dtype=float)
    if X.ndim == 1:
        is_many = False
        X = X[None, :]
    else:
        is_many = True

    if a is None:
        a = -1
        if skip_out is None:
            skip_out = False

    if b is None:
        b = 1
        if skip_out is None:
            skip_out = False

    if skip_out is None:
        skip_out = True

    d = len(A)
    n = teneva.shape(A)
    m = X.shape[0]
    a, b, n = teneva.grid_prep_opts(a, b, n, d)

    if funcs is None:
        def gen_def_func(ai, bi, ni):
            return lambda x: func_basis(teneva.poi_scale(x, ai, bi, 'cheb'), ni)

        funcs = [gen_def_func(ai, bi, ni) for ai, bi, ni in zip(a, b, n)]

    try:
        msg = 'Number of functions must be the same as TT-dimension'
        assert len(funcs) == d, msg
    except TypeError:
        funcs = [funcs] * d

    T = [f(x).T[:, :ni] for f, x, ni in zip(funcs, X.T, n)]
    y = np.ones(m) * z

    for i in range(m):
        if skip_out:
            if np.max(a - X[i, :]) > 1.E-99 or np.max(X[i, :] - b) > 1.E-99:
                continue

        Q = np.einsum('rjq,j->rq', A[0], T[0][i])
        for j in range(1, d):
            Q = Q @ np.einsum('rjq,j->rq', A[j], T[j][i])
        y[i] = Q[0, 0]

    return y if is_many else y[0]


def func_get_spectral(Y, X, H):
    """Draft. TODO"""
    X = np.asanyarray(X)

    assert X.ndim == 1 and len(X) == len(Y)

    def f(v, arg):
        c, y = arg
        return np.einsum("i,j,ijk->k", v, y[:c.shape[1]], c)

    return reduce(f, zip(Y, H(X)), np.array([1]))[0]


def func_gets(A, m=None, kind='cheb'):
    """Compute the functional approximation (TT-tensor) all over the new grid.

    Args:
        A (list): TT-tensor of the interpolation coefficients (it has d
            dimensions).
        m (int, float, list, np.ndarray): tensor size for each dimension of the
            new grid (list or np.ndarray of length d). It may be also
            int/float, then the size for each dimension will be the same. If it
            is not set, then original grid size (from the interpolation) will be
            used.
        kind (str): kind of the basis ("cheb" or "sin").

    Returns:
        list: TT-tensor of the approximated function values on the full new
        grid. This relates to the d-dimensional array of the shape m.

    Note:
        Sometimes additional rounding of the result is relevant. Use for this
        Z = teneva.truncate(Z, e) (e.g., e = 1.E-8) after the function call.

    """
    assert kind in ['cheb', 'sin']

    d = len(A)
    n = teneva.shape(A)
    m = n if m is None else teneva.grid_prep_opt(m, d, int)

    Z = []
    for k in range(d):
        if kind == 'cheb':
            I = np.arange(m[k], dtype=int).reshape((-1, 1))
            X = teneva.ind_to_poi(I, -1., +1., m[k], 'cheb').reshape(-1)
            T = func_basis(X, n[k])
        elif kind == 'sin':
            X = np.linspace(0, np.pi, m[k] + 2)[1:-1]
            T = np.sin(np.outer(X, np.arange(1, n[k] + 1))).T
        Z.append(np.einsum('riq,ij->rjq', A[k], T))

    return Z


def func_int(Y, kind='cheb'):
    """Compute the TT-tensor for functional interpolation coefficients.

    Args:
        Y (list): TT-tensor with function values on the corresponding grid.

    Returns:
        list: TT-tensor that collects interpolation coefficients. It has the
        same shape as the given tensor Y.

    Note:
        Sometimes additional rounding of the result is relevant. Use for this
        A = teneva.truncate(A, e) (e.g., e = 1.E-8) after the function call.

    """
    assert kind in ['cheb', 'sin']

    A = [None] * len(Y)

    for k, y in enumerate(Y):
        if kind == 'cheb':
            A[k] = dct(y, 1, axis=1) / (y.shape[1] - 1)
            A[k][:, 0, :] /= 2.
            A[k][:, -1, :] /= 2.
        elif kind == 'sin':
            A[k] = dst(y, 1, axis=1) / (y.shape[1] + 1)

    return A


def func_int_general(Y, X, basis_func, rcond=1.E-6):
    """Compute the TT-tensor for functional interpolation coefficients.

    Args:
        Y (list): TT-tensor.
        X (list, np.ndarray): values of continuous argument for each TT-core.
            It should be 2-dim array or 1-dim array if the values are the same
            for all TT-cores.
        basis_func (function): function, which corresponds to the values of
            the basis functions in the TT-format. It should return np.ndarray
            of the size n x m, where n is a number of one-dimensional points
            and m is a number of basis functions.

    Returns:
        list: TT-tensor, which represents the interpolation coefficients in
        terms of the functional TT-format.

    """
    d = len(Y)
    X = np.asarray(X)

    if X.ndim == 1:
        H_mats = [basis_func(X).T] * d
        X = [X] * d
    else:
        H_mats = [None] * d

    A = []
    for G, X_curr, H_mat in zip(Y, X, H_mats):
        if H_mat is None:
            H_mat = basis_func(X_curr).T

        r1, n, r2 = G.shape
        M = np.transpose(G, [1, 0, 2]).reshape(n, -1)

        Q = sp.linalg.lstsq(H_mat, M, overwrite_a=False, overwrite_b=True,
            rcond=rcond)[0]
        Q = np.transpose(Q.reshape(n, r1, r2), [1, 0, 2])
        A.append(Q)

    return A


def func_sum(A, a, b, kind='cheb'):
    """Integrate a function from its functional approximation in the TT-format.

    Args:
        A (list): TT-tensor of the interpolation coefficients (it has d
            dimensions).
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.
        kind (str): kind of the basis ("cheb" or "sin").

    Returns:
        float: the value of the integral.

    Note:
        This function works only for symmetric grids!

    """
    assert kind in ['cheb', 'sin']

    d = len(A)
    n = teneva.shape(A)
    n_max = max(n)
    a, b, n = teneva.grid_prep_opts(a, b, n, d)

    if kind == 'cheb':
        p = 2. / (1 - np.arange(0, n_max, 2)**2)
    elif kind == 'sin':
        p = 2. / np.arange(1, n_max + 1, 2)

    v = np.array([[1.]])
    for ak, bk, y, nk in zip(a, b, A, n):
        v = v @ (p[:(nk + 1)//2] @ y[:, ::2])
        v *= (bk - ak) / 2.

    return v.item()



"""Package teneva, module props: various properties of TT-tensors.

This module contains the basic properties of TT-tensors, including "erank",
"ranks", "shape", etc.

"""
import numpy as np


def erank(Y):
    """Compute effective TT-rank of the given TT-tensor.

    Effective TT-rank r of a TT-tensor Y with shape [n_1, n_2, ..., n_d]
    and TT-ranks r_0, r_1, ..., r_d (r_0 = r_d = 1) is a solution of
    equation n_1 r + \sum_{\alpha=2}^{d-1} n_\alpha r^2 + n_d r =
    \sum_{\alpha=1}^{d} n_\alpha r_{\alpha-1} r_{\alpha}.

    The representation with a constant TT-rank r (r_0 = 1, r_1 = r_2 = ... =
    r_{d-1} = r, r_d = 1) yields the same total number of parameters as in the
    original decomposition of the tensor Y.

    Args:
        Y (list): TT-tensor.

    Returns:
        float: effective TT-rank.

    """
    d, n, r = len(Y), shape(Y), ranks(Y)

    if d == 2:
        return r[1]

    sz = np.dot(n * r[0:d], r[1:])
    b = r[0] * n[0] + n[d-1] * r[d]
    a = np.sum(n[1:d-1])

    return (np.sqrt(b * b + 4 * a * sz) - b) / (2 * a)


def ranks(Y):
    """Return the TT-ranks of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    Returns:
        np.ndarray: TT-ranks in form of the 1D array of ints of the length
        d+1, where d is a number of tensor dimensions (the first and last
        elements are equal 1).

    """
    return np.array([1] + [G.shape[2] for G in Y], dtype=int)


def shape(Y):
    """Return the shape of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    Returns:
        np.ndarray: shape of the tensor in form of the 1D array of ints of the
        length d, where d is a number of tensor dimensions.

    """
    return np.array([G.shape[1] for G in Y], dtype=int)


def size(Y):
    """Return the size (number of parameters) of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    Returns:
        int: total number of parameters in the TT-representation (it is a sum
        of sizes of all TT-cores).

    """
    return np.sum([G.size for G in Y])



"""Package teneva, module utils: various (inner) helper functions.

This module contains some helper functions which used in other modules.

"""
import numpy as np
import teneva
from time import perf_counter as tpc


def _info_appr(info, t, nswp, e, e_vld, log=False):
    info['t'] = tpc() - t

    if info['stop'] is None:
        if e_vld is not None and info['e_vld'] >= 0:
            if info['e_vld'] <= e_vld and not np.isinf(info['e_vld']):
                info['stop'] = 'e_vld'

    if info['stop'] is None:
        if e is not None and info['e'] >= 0:
            if info['e'] <= e and not np.isinf(info['e']):
                info['stop'] = 'e'

    if info['stop'] is None:
        if nswp is not None:
            if info['nswp'] >= nswp:
                info['stop'] = 'nswp'

    if log:
        text = ''

        if info['nswp'] == 0:
            text += f'# pre | '
        else:
            text += f'# {info["nswp"]:-3d} | '

        text += f'time: {info["t"]:-10.3f} | '

        if 'm' in info:
            if 'with_cache' in info and info['with_cache']:
                text += f'evals: {info["m"]:-8.2e} '
                text += f'(+ {info["m_cache"]:-8.2e}) | '
            else:
                text += f'evals: {info["m"]:-8.2e} | '

        text += f'rank: {info["r"]:-5.1f} | '

        if info['e_vld'] >= 0:
            text += f'e_vld: {info["e_vld"]:-7.1e} | '

        if info['e'] >= 0:
            text += f'e: {info["e"]:-7.1e} | '

        if info['stop']:
            text += f'stop: {info["stop"]} | '

        print(text)

    return info['stop']


def _is_num(A):
    return isinstance(A, (int, float))


def _maxvol(A, tau=1.1, dr_min=0, dr_max=0, tau0=1.05, k0=100):
    n, r = A.shape
    dr_max = min(dr_max, n - r)
    dr_min = min(dr_min, dr_max)

    if n <= r:
        I = np.arange(n, dtype=int)
        B = np.eye(n, dtype=float)
    elif dr_max == 0:
        I, B = teneva.maxvol(A, tau0, k0)
    else:
        I, B = teneva.maxvol_rect(A, tau, dr_min, dr_max, tau0, k0)

    return I, B


def _ones(k, m=1):
    return np.ones((k, m), dtype=int)


def _rand(seed=None):
    if seed is None or isinstance(seed, int):
        return np.random.default_rng(seed)
    else:
        return seed


def _range(n):
    return np.arange(n).reshape(-1, 1)


def _reshape(A, n, order='F'):
    return np.reshape(A, n, order=order)


def _vector_index_expand(q, i):
    if i < 0:
        if i == -1:
            ind = [1] * q
        else:
            raise ValueError('Only "-1" is supported for negative indices.')
    else:
        ind = []
        for _ in range(q):
            ind.append(i % 2)
            i = int(i / 2)
        if i > 0:
            raise ValueError('Index is out of range.')

    return ind


def _vector_index_prepare(q, i):
    n = 1 << q
    if i >= n or i < -n:
        raise ValueError('Incorrect index.')
    return i if i >= 0 else n + i



"""Package teneva, module cross: construct TT-tensor, using TT-cross.

This module contains the function "cross" which computes the TT-approximation
for implicit tensor given functionally  by the rank-adaptive multidimensional
cross approximation method in the TT-format (TT-cross).

"""
import numpy as np
import teneva
from time import perf_counter as tpc


def cross(f, Y0, m=None, e=None, nswp=None, tau=1.1, dr_min=1, dr_max=1,
          tau0=1.05, k0=100, info={}, cache=None, I_vld=None, y_vld=None,
          e_vld=None, cb=None, func=None, m_cache_scale=5, log=False):
    """Compute the TT-approximation for implicit tensor given functionally.

    This function computes the TT-approximation for implicit tensor given
    functionally by the rank-adaptive multidimensional cross approximation
    method in the TT-format (TT-cross).

    Args:
        f (function): function f(I) which computes tensor elements for the
            given set of multi-indices I, where I is a 2D np.ndarray of the
            shape [samples, dimensions]. The function should return 1D
            np.ndarray of the length equals to samples, which relates to the
            values of the target function for all provided samples. If the
            function returns None, then the algorithm will be interrupted and
            the current result will be returned  (in the info dictionary, in
            this case, the stop type of the algorithm will be "func").
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        m (int): optional limit on the maximum number of requests to the
            objective function (> 0). If specified, then the total number of
            requests will not exceed this value. Note that the actual number of
            requests may be less, since the values are requested in batches.
        e (float): optional algorithm convergence criterion (> 0). If between
            iterations the relative rate of solution change is less than this
            value, then the operation of the algorithm will be interrupted.
        nswp (int): optional maximum number of iterations (sweeps) of the
            algorithm (>= 0). One sweep corresponds to a complete pass of all
            tensor TT-cores from left to right and then from right to left. If
            nswp = 0, then only "maxvol-preiteration" will be performed.
        tau (float): accuracy parameter (>= 1) for the algorithm "maxvol_rect"
            (see "maxvol_rect" function for more details).
        dr_min (int): minimum number of added rows in the process of adaptively
            increasing the TT-rank of the approximation using the algorithm
            "maxvol_rect" (see "maxvol_rect" function for more details). Note
            that dr_min should be no bigger than dr_max.
        dr_max (int): maximum number of added rows in the process of adaptively
            increasing the TT-rank of the approximation using the algorithm
            "maxvol_rect" (see "maxvol_rect" function for more details). Note
            that dr_max should be no less than dr_min. If dr_max = 0,
            then basic maxvol algorithm will be used (rank will be constant).
        tau0 (float): accuracy parameter (>= 1) for the algorithm "maxvol" (see
            "maxvol" function for more details). It will be used while maxvol
            preiterations and while the calls of "maxvol" function from the
            "maxvol_rect" algorithm.
        k0 (int): maximum number of maxvol iterations (>= 1; see "maxvol"
            function for more details). It will be used while maxvol
            preiterations and while the calls of "maxvol" function from the
            "maxvol_rect" algorithm.
        info (dict): an optionally set dictionary, which will be filled with
            reference information about the process of the algorithm operation.
            At the end of the function work, it will contain parameters: m -
            total number of requests to the target function; e - the final
            value of the convergence criterion; e_vld - the final error on the
            validation dataset; nswp - the real number of performed
            iterations (sweeps); m_cache - total number of requests to the
            cache; stop - stop type of the algorithm (see note below).
        cache (dict): an optionally set dictionary, which will be filled with
            requested function values. Since the algorithm sometimes requests
            the same tensor indices, the use of such a cache may speed up the
            operation of the algorithm if the time to find a value in the cache
            is less than the time to calculate the target function.
        I_vld (np.ndarray): optional multi-indices for items of validation
            dataset in the form of array of the shape [samples_vld, d], where
            samples_vld is a size of the validation dataset.
        y_vld (np.ndarray): optional values of the tensor for multi-indices
            I_vld of validation dataset in the form of array of the shape
            [samples_vld].
        e_vld (float): optional algorithm convergence criterion (> 0). If
            after sweep, the error on the validation dataset is less than this
            value, then the operation of the algorithm will be interrupted.
        cb (function): optional callback function. It will be called after
            every sweep and the accuracy check with the arguments: Y,
            info and opts, where Y is the current approximation (TT-tensor),
            info is the info dictionary and the dictionary opts contains fields
            Ir, Ic, cache and Yold. If the callback returns a true value, then
            the algorithm will be stopped (in the info dictionary, in this case,
            the stop type of the algorithm will be "cb").
        func (function): if this function is set, then it will replace the inner
            function _func, which deals with requests to the objective
            function f. This argument is used only for internal experiments.
        m_cache_scale (int): if the number of requests to the cache is
            m_cache_scale times greater than the current number of requests to
            the target function, then the algorithm will stop (in the info
            dictionary, in this case, the stop type will be "conv").
        log (bool): if flag is set, then the information about the progress of
            the algorithm will be printed after each sweep.

    Returns:
        list: TT-Tensor which approximates the implicit tensor.

    Note:
        Note that at list one of the arguments m / e / nswp / e_vld should be
        set by user. The end of the algorithm operation occurs when one of the
        following criteria is reached: 1) the target function returns None
        instead of a batch of values; 2) the maximum allowable number of the
        objective function calls (m) has been done (more precisely, if the next
        request will result in exceeding this value, then algorithm will not
        perform this new request); 3) the convergence criterion (e) is reached;
        4) the maximum number of iterations (nswp) is performed; 5) the
        algorithm is already converged (all requested values are in the cache
        already); 6) the error on validation dataset I_vld, y_vld is less than
        e_vld; 7) the callback function returns true value. The related stop
        type (func, m, e, nswp, conv, e_vld or cb) will be written into the
        item stop of the info dictionary.

        The resulting TT-tensor usually has overestimated ranks, so you should
        truncate the result. Use for this Y = teneva.truncate(Y, e) (e.g.,
        e = 1.E-8) after this function call.

    """
    if m is None and e is None and nswp is None:
        if I_vld is None or y_vld is None:
            raise ValueError('One of arguments m/e/nswp should be set')
        elif e_vld is None:
            raise ValueError('One of arguments m/e/e_vld/nswp should be set')
    if e_vld is not None and (I_vld is None or y_vld is None):
        raise ValueError('Validation dataset is not set')

    _time = tpc()
    info.update({'r': teneva.erank(Y0), 'e': -1, 'e_vld': -1, 'nswp': 0,
        'stop': None, 'm': 0, 'm_cache': 0, 'm_max': int(m) if m else None,
        'with_cache': cache is not None})

    d = len(Y0)
    n = teneva.shape(Y0)
    Y = teneva.copy(Y0)

    Ig = [teneva._reshape(np.arange(k, dtype=int), (-1, 1)) for k in n]
    Ir = [None for i in range(d+1)]
    Ic = [None for i in range(d+1)]

    R = np.ones((1, 1))
    for i in range(d):
        G = np.tensordot(R, Y[i], 1)
        Y[i], R, Ir[i+1] = _iter(G, Ig[i], Ir[i], tau0=tau0, k0=k0, ltr=True)
    Y[d-1] = np.tensordot(Y[d-1], R, 1)

    R = np.ones((1, 1))
    for i in range(d-1, -1, -1):
        G = np.tensordot(Y[i], R, 1)
        Y[i], R, Ic[i] = _iter(G, Ig[i], Ic[i+1], tau0=tau0, k0=k0, ltr=False)
    Y[0] = np.tensordot(R, Y[0], 1)

    info['e_vld'] = teneva.accuracy_on_data(Y, I_vld, y_vld)
    teneva._info_appr(info, _time, nswp, e, e_vld, log)

    while True:
        Yold = teneva.copy(Y)

        R = np.ones((1, 1))
        for i in range(d):
            Z = (func or _func)(f, Ig[i], Ir[i], Ic[i+1], info, cache)
            if info['stop']:
                Y[i] = np.tensordot(R, Y[i], 1)
                info['r'] = teneva.erank(Y)
                info['e'] = teneva.accuracy(Y, Yold)
                info['e_vld'] = teneva.accuracy_on_data(Y, I_vld, y_vld)
                teneva._info_appr(info, _time, nswp, e, e_vld, log)
                return Y
            Y[i], R, Ir[i+1] = _iter(Z, Ig[i], Ir[i],
                tau, dr_min, dr_max, tau0, k0, ltr=True)
        Y[d-1] = np.tensordot(Y[d-1], R, 1)

        R = np.ones((1, 1))
        for i in range(d-1, -1, -1):
            Z = (func or _func)(f, Ig[i], Ir[i], Ic[i+1], info, cache)
            if info['stop']:
                Y[i] = np.tensordot(Y[i], R, 1)
                info['r'] = teneva.erank(Y)
                info['e'] = teneva.accuracy(Y, Yold)
                info['e_vld'] = teneva.accuracy_on_data(Y, I_vld, y_vld)
                teneva._info_appr(info, _time, nswp, e, e_vld, log)
                return Y
            Y[i], R, Ic[i] = _iter(Z, Ig[i], Ic[i+1],
                tau, dr_min, dr_max, tau0, k0, ltr=False)
        Y[0] = np.tensordot(R, Y[0], 1)

        info['nswp'] += 1
        info['r'] = teneva.erank(Y)
        info['e'] = teneva.accuracy(Y, Yold)
        info['e_vld'] = teneva.accuracy_on_data(Y, I_vld, y_vld)

        if info['m_cache'] > m_cache_scale * info['m']:
            info['stop'] = 'conv'

        if cb:
            opts = {'Yold': Yold, 'Ir': Ir, 'Ic': Ic, 'cache': cache}
            if cb(Y, info, opts) is True:
                info['stop'] = info['stop'] or 'cb'

        if teneva._info_appr(info, _time, nswp, e, e_vld, log):
            return Y


def _func(f, Ig, Ir, Ic, info, cache=None):
    n = Ig.shape[0]
    r1 = Ir.shape[0] if Ir is not None else 1
    r2 = Ic.shape[0] if Ic is not None else 1

    I = np.kron(np.kron(teneva._ones(r2), Ig), teneva._ones(r1))

    if Ir is not None:
        Ir_ = np.kron(teneva._ones(n * r2), Ir)
        I = np.hstack((Ir_, I))

    if Ic is not None:
        Ic_ = np.kron(Ic, teneva._ones(r1 * n))
        I = np.hstack((I, Ic_))

    y = _func_eval(f, I, info, cache)

    if y is not None:
        return teneva._reshape(y, (r1, n, r2))


def _func_eval(f, I, info, cache=None):
    if cache is None:
        if info['m_max'] is not None and info['m'] + len(I) > info['m_max']:
            info['stop'] = 'm'
            return
        y = f(I)
        if y is None:
            info['stop'] = 'func'
            return
        info['m'] += len(I)
        return np.array(y, dtype=float)

    I_new = np.array([i for i in I if tuple(i) not in cache])
    if len(I_new):
        if info['m_max'] is not None and info['m'] + len(I_new) > info['m_max']:
            info['stop'] = 'm'
            return
        y_new = f(I_new)
        if y_new is None:
            info['stop'] = 'func'
            return
        for k, i in enumerate(I_new):
            cache[tuple(i)] = float(y_new[k])

    info['m'] += len(I_new)
    info['m_cache'] += len(I) - len(I_new)

    return np.array([cache[tuple(i)] for i in I], dtype=float)


def _iter(Z, Ig, I, tau=1.1, dr_min=0, dr_max=0, tau0=1.05, k0=100, ltr=True):
    r1, n, r2 = Z.shape

    if ltr:
        Z = teneva._reshape(Z, (r1 * n, r2))
    else:
        Z = teneva._reshape(Z, (r1, n * r2)).T

    Q, R = np.linalg.qr(Z)

    ind, B = teneva._maxvol(Q, tau, dr_min, dr_max, tau0, k0)

    if ltr:
        G = teneva._reshape(B, (r1, n, -1))
        R = Q[ind, :] @ R
        I_new = np.kron(Ig, teneva._ones(r1))
        if I is not None:
            I_old = np.kron(teneva._ones(n), I)
            I_new = np.hstack((I_old, I_new))

    else:
        G = teneva._reshape(B.T, (-1, n, r2))
        R = (Q[ind, :] @ R).T
        I_new = np.kron(teneva._ones(r2), Ig)
        if I is not None:
            I_old = np.kron(I, teneva._ones(n))
            I_new = np.hstack((I_new, I_old))

    return G, R, I_new[ind, :]



__version__ = '0.14.8'


from .act_many import add_many
from .act_many import outer_many


from .act_one import copy
from .act_one import interface
from .act_one import get
from .act_one import get_and_grad
from .act_one import get_many
from .act_one import getter
from .act_one import mean
from .act_one import norm
from .act_one import qtt_to_tt
from .act_one import sum
from .act_one import tt_to_qtt


from .act_two import accuracy
from .act_two import add
from .act_two import mul
from .act_two import mul_scalar
from .act_two import outer
from .act_two import sub


from .als import als


from .als_func import als_func


from .anova import ANOVA
from .anova import anova


from .anova_func import ANOVA_func
from .anova_func import anova_func


from .core import core_dot
from .core import core_dot_inv
from .core import core_dot_maxvol
from .core import core_qr_rand
from .core import core_qtt_to_tt
from .core import core_stab
from .core import core_tt_to_qtt


from .cross import cross


from .cross_act import cross_act


from .data import accuracy_on_data
from .data import cache_to_data


from .func import func_basis
from .func import func_diff_matrix
from .func import func_diff_matrix_apply
from .func import func_get
from .func import func_gets
from .func import func_int
from .func import func_int_general
from .func import func_sum


from .func_full import func_get_full
from .func_full import func_gets_full
from .func_full import func_int_full
from .func_full import func_sum_full


from .grid import grid_flat
from .grid import grid_prep_opt
from .grid import grid_prep_opts
from .grid import ind_qtt_to_tt
from .grid import ind_to_poi
from .grid import ind_tt_to_qtt
from .grid import poi_scale
from .grid import poi_to_ind


from .matrices import matrix_delta


from .maxvol import maxvol
from .maxvol import maxvol_rect


from .optima import optima_qtt
from .optima import optima_tt
from .optima import optima_tt_beam
from .optima import optima_tt_max


from .optima_func import optima_func_tt_beam


from .props import erank
from .props import ranks
from .props import shape
from .props import size


from .sample import sample
from .sample import sample_square
from .sample import sample_lhs
from .sample import sample_rand
from .sample import sample_rand_poi
from .sample import sample_tt


from .sample_func import sample_func


from .stat import cdf_confidence
from .stat import cdf_getter


from .svd import matrix_skeleton
from .svd import matrix_svd
from .svd import svd
from .svd import svd_matrix
from .svd import svd_incomplete


from .tensors import const
from .tensors import delta
from .tensors import poly
from .tensors import rand
from .tensors import rand_custom
from .tensors import rand_norm
from .tensors import rand_stab


from .transformation import full
from .transformation import full_matrix
from .transformation import orthogonalize
from .transformation import orthogonalize_left
from .transformation import orthogonalize_right
from .transformation import truncate


from .utils import _info_appr
from .utils import _is_num
from .utils import _maxvol
from .utils import _ones
from .utils import _rand
from .utils import _range
from .utils import _reshape
from .utils import _vector_index_expand
from .utils import _vector_index_prepare


from .vectors import vector_delta


from .vis import show



"""Package teneva, module func_full: functional full format with interpolation.

This module contains the functions for construction of the functional
representation in the full format with Chebyshev interpolation, as well as
calculating the values of the function using the constructed interpolation
coefficients. See module "func" with the same functions in the TT-format. The
functions presented in this module are especially relevant for the
one-dimensional (and two-dimensional) case, when the TT-decomposition cannot be
applied.

"""
import numpy as np
import teneva


def func_get_full(X, A, a, b, z=0., skip_out=True):
    """Compute the Chebyshev approximation in given points (approx. f(X)).

    Args:
        X (np.ndarray): spatial points of interest (it is 2D array of the shape
            [samples, d], where d is the number of dimensions).
        A (list): TT-tensor of the interpolation coefficients (it has d
            dimensions).
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.
        z (float): the value for points, which are outside the spatial grid.
        skip_out (bool): if flag is set, then the values outside the spatial
            grid will be set to z values.

    Returns:
        np.ndarray: approximated function values in X points (it is 1D array of
        the shape [samples]).

    Note:
        See also the same function ("func_get") in the TT-format.

    """
    d = len(A.shape)
    n = A.shape
    m = X.shape[0]
    a, b, n = teneva.grid_prep_opts(a, b, n, d)

    # TODO: check if this operation is effective. It may be more profitable to
    # generate polynomials for each tensor mode separately:
    T = teneva.func_basis(teneva.poi_scale(X, a, b, 'cheb'), max(n))

    Y = np.ones(m) * z
    for i in range(m):
        if skip_out:
            if np.max(a - X[i, :]) > 1.E-99 or np.max(X[i, :] - b) > 1.E-99:
                # We skip the points outside the grid bounds:
                continue

        Q = A.copy()
        for j in range(d):
            Q = np.tensordot(Q, T[:n[j], i, j], axes=([0], [0]))
        Y[i] = Q

    return Y


def func_gets_full(A, a, b, m=None):
    """Compute the Chebyshev approximation all over the new grid.

    Args:
        A (np.ndarray): d-dimensional tensor of the interpolation coefficients.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.
        m (int, float, list, np.ndarray): tensor size for each dimension of the
            new grid (list or np.ndarray of length d). It may be also
            int/float, then the size for each dimension will be the same. If it
            is not set, then original grid size (from the interpolation) will be
            used.

    Returns:
        np.ndarray: array of the approximated function values on the full new
        grid. This is the d-dimensional array of the shape m.

    Note:
        This function is not efficient in the full format, and corresponds to a
        simple explicit calculation of all values at the nodes of the new grid.
        At the same time, in the TT-format ("func_gets"), the corresponding
        operation is carried out in an implicit efficient form. Accordingly,
        this function is provided only for uniformity of presentation.

    """
    d = len(A.shape)
    n = A.shape
    m = n if m is None else teneva.grid_prep_opt(m, d, int)

    I = teneva.grid_flat(m)
    X = teneva.ind_to_poi(I, -1., +1., m, 'cheb')
    Z = func_get_full(X, A, -1., +1.)
    Z = Z.reshape(m, order='F')

    return Z


def func_int_full(Y):
    """Compute the tensor for Chebyshev interpolation coefficients.

    Args:
        Y (np.ndarray): d-dimensional array with function values on the
            Chebyshev grid.

    Returns:
        np.ndarray: array that collects interpolation coefficients. It has the
        same shape as the given tensor Y.

    Note:
        See also the same function ("func_int") in the TT-format.

    """
    d = len(Y.shape)
    A = Y.copy()
    for k in range(d):
        n = np.array(Y.shape, dtype=int)
        m = n[k]
        n[[0, k]] = n[[k, 0]]
        A = np.swapaxes(A, 0, k)
        A = A.reshape((m, -1), order='F')
        A = np.vstack([A, A[m-2 : 0 : -1, :]])
        A = np.fft.fft(A, axis=0).real
        A = A[:m, :] / (m - 1)
        A[0, :] /= 2.
        A[m-1, :] /= 2.
        A = A.reshape(n, order='F')
        A = np.swapaxes(A, 0, k)
    return A


def func_sum_full(A, a, b):
    """Integrate the function from its Chebyshev approximation.

    Args:
        A (np.ndarray): d-dimensional tensor of the interpolation coefficients.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.

    Returns:
        float: the value of the integral.

    Note:
        This function works only for symmetric grids! See also the same
        function ("func_sum") in the TT-format.

    """
    d = len(A.shape)
    n = A.shape
    a, b, n = teneva.grid_prep_opts(a, b, n, d)

    for k in range(d):
        if abs(abs(b[k]) - abs(a[k])) > 1.E-16:
            raise ValueError('This function works only for symmetric grids')

    v = A.copy()
    for k in range(d):
        v = v.reshape(n[k], -1)
        p = np.arange(n[k])[::2]
        p = np.repeat(p.reshape(-1, 1), v.shape[1], axis=1)
        v = np.sum(v[::2, :] * 2. / (1. - p**2), axis=0)
        v *= (b[k] - a[k]) / 2.

    return v[0]



import numpy as np
from scipy.optimize import rosen
import teneva
from time import perf_counter as tpc
import unittest


class TestFuncFuncGet(unittest.TestCase):
    def setUp(self):
        self.eps = 1.E-10

    def test_base(self):
        f = lambda X: rosen(X.T)        # Target function
        f_grid = lambda I: f(teneva.ind_to_poi(I, a, b, n, 'cheb'))
        a = [-2., -4., -3., -2.]        # Grid lower bounds
        b = [+2., +3., +4., +2.]        # Grid upper bounds
        n = [5, 6, 7, 8]                # Grid size
        Y0 = teneva.rand(n, r=2)        # Initial approximation for TT-cross
        e = 1.E-3                       # Accuracy for TT-CROSS
        eps = 1.E-6                     # Accuracy for truncation
        Y = teneva.cross(f_grid, Y0, e=e)
        Y = teneva.truncate(Y, eps)

        A = teneva.func_int(Y)

        X = np.array([
            [0., 0., 0., 0.],
            [0., 2., 3., 2.],
            [1., 1., 1., 1.],
        ])

        Z = teneva.func_get(X, A, a, b)

        err = np.linalg.norm(Z - f(X))
        self.assertLess(err, self.eps)

        x = X[1]
        z = teneva.func_get(x, A, a, b)
        err = np.linalg.norm(z - f(x))
        self.assertLess(err, self.eps)


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



import numpy as np
import teneva
from time import perf_counter as tpc
import unittest


class TestActOneCopy(unittest.TestCase):
    def setUp(self):
        self.eps = 1.E-16

    def test_array(self):
        Y = np.random.randn(3, 4, 5)
        Z = teneva.copy(Y)

        y1 = Y[1, 2, 0]
        Z[1, 2, 0] = 42.
        y2 = Y[1, 2, 0]

        e = np.abs(y1-y2)
        self.assertLess(e, self.eps)

    def test_base(self):
        Y = teneva.rand([10] * 5, r=3, seed=42)
        Z = teneva.copy(Y)

        y1 = Y[2][1, 2, 0]
        Z[2][1, 2, 0] = 42.
        y2 = Y[2][1, 2, 0]

        e = np.abs(y1-y2)
        self.assertLess(e, self.eps)

    def test_base_rev(self):
        Y = teneva.rand([10] * 5, r=3, seed=42)
        Z = teneva.copy(Y)

        z1 = Z[2][1, 2, 0]
        Y[2][1, 2, 0] = 42.
        z2 = Z[2][1, 2, 0]

        e = np.abs(z1-z2)
        self.assertLess(e, self.eps)

    def test_none(self):
        y = None
        z = teneva.copy(y)

        self.assertIsNone(z)

    def test_number(self):
        y = 42.
        z = teneva.copy(y)

        e = np.abs(z-y)
        self.assertLess(e, self.eps)


class TestActOneGet(unittest.TestCase):
    def setUp(self):
        self.Y = teneva.rand([10] * 5, r=3, seed=42)
        self.Z = teneva.full(self.Y)
        self.i = [1, 2, 3, 4, 5]
        self.eps = 1.E-16

    def test_base(self):
        i = np.array(self.i)
        y = teneva.get(self.Y, i)
        z = self.Z[tuple(i)]
        e = np.abs(y-z)
        self.assertLess(e, self.eps)

    def test_list(self):
        i = self.i
        y = teneva.get(self.Y, i)
        z = self.Z[tuple(i)]
        e = np.abs(y-z)
        self.assertLess(e, self.eps)

    def test_many(self):
        i = np.array(self.i)
        I = [i, i-1, i+1]
        y = teneva.get(self.Y, I)
        z = np.array([self.Z[tuple(i)] for i in I])
        e = np.abs(np.max(y-z))
        self.assertLess(e, self.eps)


class TestActOneGetAndGrad(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [10] * self.d
        self.Y = teneva.rand(self.n, r=3, seed=42)
        self.i = [1, 2, 3, 4, 5]
        self.lr = 1.E-4
        self.eps = 1.E-16

    def test_base(self):
        y, dY = teneva.get_and_grad(self.Y, self.i)

        Z = teneva.copy(self.Y)
        for k in range(self.d):
            Z[k] -= self.lr * dY[k]

        z = teneva.get(Z, self.i)
        e = teneva.accuracy(self.Y, Z)

        self.assertLess(z, y)
        self.assertLess(e, self.lr)

    def test_get(self):
        y, dY = teneva.get_and_grad(self.Y, self.i)
        z = teneva.get(self.Y, self.i)
        e = np.abs(y-z)
        self.assertLess(e, self.eps)


class TestActOneGetMany(unittest.TestCase):
    def setUp(self):
        self.Y = teneva.rand([10] * 5, r=3, seed=42)
        self.Z = teneva.full(self.Y)
        self.I = [
            [1, 2, 3, 4, 5],
            [5, 4, 3, 2, 1],
            [1, 2, 3, 2, 1],
        ]
        self.eps = 1.E-16

    def test_base(self):
        I = np.array(self.I)
        y = teneva.get_many(self.Y, I)
        z = np.array([self.Z[tuple(i)] for i in I])
        e = np.abs(np.max(y-z))
        self.assertLess(e, self.eps)

    def test_list(self):
        I = self.I
        y = teneva.get_many(self.Y, I)
        z = np.array([self.Z[tuple(i)] for i in I])
        e = np.abs(np.max(y-z))
        self.assertLess(e, self.eps)


class TestActOneInterface(unittest.TestCase):
    def setUp(self):
        self.n = [6, 5, 4, 3, 2]
        self.d = len(self.n)
        self.Y = teneva.rand(self.n, r=3, seed=42)
        self.i = [5, 4, 3, 2, 1]
        self.P = [
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
            [0.1, 0.2, 0.3, 0.4, 0.5],
            [0.1, 0.2, 0.3, 0.4],
            [0.1, 0.2, 0.3],
            [0.1, 0.2]]

        self.d2 = 8
        self.n2 = [5] * self.d2
        self.Y2 = teneva.rand(self.n2, r=4, seed=42)
        self.p2 = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

        self.eps = 1.E-15

    def test_base(self):
        phi_r = teneva.interface(self.Y)
        phi_l = teneva.interface(self.Y, ltr=True)

        self.assertEqual(len(phi_r[0]), 1)
        self.assertEqual(len(phi_r[-1]), 1)
        self.assertEqual(len(phi_l[0]), 1)
        self.assertEqual(len(phi_l[-1]), 1)

        v_r = phi_r[0].item()
        v_l = phi_l[-1].item()
        e = np.abs(v_r-v_l)
        self.assertLess(e, self.eps)

    def test_i(self):
        phi_r = teneva.interface(self.Y, i=self.i)
        phi_l = teneva.interface(self.Y, i=self.i, ltr=True)

        self.assertEqual(len(phi_r[0]), 1)
        self.assertEqual(len(phi_r[-1]), 1)
        self.assertEqual(len(phi_l[0]), 1)
        self.assertEqual(len(phi_l[-1]), 1)

        v_r = phi_r[0].item()
        v_l = phi_l[-1].item()
        e = np.abs(v_r-v_l)
        self.assertLess(e, self.eps)

    def test_i_p(self):
        phi_r = teneva.interface(self.Y, self.P, self.i)
        phi_l = teneva.interface(self.Y, self.P, self.i, ltr=True)

        self.assertEqual(len(phi_r[0]), 1)
        self.assertEqual(len(phi_r[-1]), 1)
        self.assertEqual(len(phi_l[0]), 1)
        self.assertEqual(len(phi_l[-1]), 1)

        v_r = phi_r[0].item()
        v_l = phi_l[-1].item()
        e = np.abs(v_r-v_l)
        self.assertLess(e, self.eps)

    def test_norm_natural(self):
        phi_r = teneva.interface(self.Y, norm='n')
        phi_l = teneva.interface(self.Y, norm='n', ltr=True)

        self.assertEqual(len(phi_r[0]), 1)
        self.assertEqual(len(phi_r[-1]), 1)
        self.assertEqual(len(phi_l[0]), 1)
        self.assertEqual(len(phi_l[-1]), 1)

        v_r = phi_r[0].item()
        v_l = phi_l[-1].item()
        e = np.abs(v_r-v_l)
        self.assertLess(e, self.eps)

    def test_norm_none(self):
        phi_r = teneva.interface(self.Y, norm=None)
        phi_l = teneva.interface(self.Y, norm=None, ltr=True)

        self.assertEqual(len(phi_r[0]), 1)
        self.assertEqual(len(phi_r[-1]), 1)
        self.assertEqual(len(phi_l[0]), 1)
        self.assertEqual(len(phi_l[-1]), 1)

        v_r = phi_r[0].item()
        v_l = phi_l[-1].item()
        e = np.abs(v_r-v_l)
        self.assertLess(e, self.eps)

    def test_p(self):
        phi_r = teneva.interface(self.Y, self.P)
        phi_l = teneva.interface(self.Y, self.P, ltr=True)

        self.assertEqual(len(phi_r[0]), 1)
        self.assertEqual(len(phi_r[-1]), 1)
        self.assertEqual(len(phi_l[0]), 1)
        self.assertEqual(len(phi_l[-1]), 1)

        v_r = phi_r[0].item()
        v_l = phi_l[-1].item()
        e = np.abs(v_r-v_l)
        self.assertLess(e, self.eps)

    def test_p_equal(self):
        phi_r = teneva.interface(self.Y2, self.p2)
        phi_l = teneva.interface(self.Y2, self.p2, ltr=True)

        self.assertEqual(len(phi_r[0]), 1)
        self.assertEqual(len(phi_r[-1]), 1)
        self.assertEqual(len(phi_l[0]), 1)
        self.assertEqual(len(phi_l[-1]), 1)

        v_r = phi_r[0].item()
        v_l = phi_l[-1].item()
        e = np.abs(v_r-v_l)
        self.assertLess(e, self.eps)


class TestActOneMean(unittest.TestCase):
    def setUp(self):
        self.n = [5] * 10
        self.Y = teneva.rand(self.n, r=3, seed=42)
        self.Z = teneva.full(self.Y)
        self.eps = 1.E-16

    def test_base(self):
        m_calc = teneva.mean(self.Y)
        m_real = np.mean(self.Z)

        e = np.abs(m_calc - m_real)
        self.assertLess(e, self.eps)

    def test_prob_zero(self):
        P = [np.zeros(k) for k in self.n]
        m = teneva.mean(self.Y, P)
        self.assertLess(np.abs(m), self.eps)


class TestActOneSum(unittest.TestCase):
    def setUp(self):
        self.n = [5] * 10
        self.Y = teneva.rand(self.n, r=3, seed=42)
        self.Z = teneva.full(self.Y)
        self.eps = 1.E-13

    def test_base(self):
        s_calc = teneva.sum(self.Y)
        s_real = np.sum(self.Z)

        e = np.abs(s_calc - s_real)
        self.assertLess(e, self.eps)


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



"""Package teneva, module matrices: various useful QTT-matrices.

This module contains the collection of functions for explicit construction of
various useful QTT-matrices (delta function and others).

"""
import numpy as np
import teneva


def matrix_delta(q, i, j, v=1.):
    """Build QTT-matrix that is zero everywhere except for a given 2D index.

    Args:
        q (int): quantization level. The resulting matrix will have the shape
            2^q x 2^q.
        i (int): the col index for nonzero element (< 2^q). Note that "negative
            index notation" is supported.
        j (int): the row index for nonzero element (< 2^q). Note that "negative
            index notation" is supported.
        v (float): the value of the matrix at index i, j.

    Returns:
        list: TT-tensor with 4D TT-cores representing the QTT-matrix.

    """
    i = teneva._vector_index_prepare(q, i)
    j = teneva._vector_index_prepare(q, j)
    ind_col = teneva._vector_index_expand(q, i)
    ind_row = teneva._vector_index_expand(q, j)
    Y = []
    for k in range(q):
        G = np.zeros((1, 2, 2, 1))
        G[0, ind_col[k], ind_row[k], 0] = 1.
        Y.append(G)
    Y[-1][0, ind_col[-1], ind_row[-1], 0] = v
    return Y



"""Package teneva, module optima_func: estimate max for function.

This module contains the novel algorithm for computation of minimum and
maximum element of the multivariate function presented as the TT-tensor
of Chebyshev coefficients.

"""
import numpy as np
import teneva


def optima_func_tt_beam(A, k, k_loc=None):
    """Find maximum modulo points in the functional TT-tensor.

    Args:
        A (list): d-dimensional TT-tensor of interpolation coefficients.
        k (int): number of selected items (candidates for the optimum) for each
            tensor mode.
        k_loc (int): optional number of local maximum to take.

    Returns:
        np.ndarray: the set of k_loc best multidimensional points (array of the
        shape [k, d]).

    """
    if k_loc is None:
        k_loc = k

    A = teneva.copy(A)
    for G in A:
        G[:, 0, :] *= np.sqrt(2.)
    A = teneva.orthogonalize(A, 0)

    X_prev = None # x by H matrix of basis functions values
    G_prev = np.array([[1.]])
    for G in A:
        X_prev, G_prev = _step_top_k(X_prev, G_prev, G, k, k_loc)

    return X_prev


def _cheb_my_poly(X, n):
    X = np.asarray(X)
    to_red = False
    if X.ndim == 0:
        X = X[None]
        to_red = True

    res = np.ones([X.shape[0], n])
    if n > 1:
        res[:, 1] = X

    for i in range(2, n):
        res[:, i] = 2*X*res[:, i - 1] - res[:, i - 2]

    res[:, 0] = np.sqrt(0.5)

    return res[0] if to_red else res


def _find_poly_max(p, clip=[-1, 1], cheb=True, take_abs=True, k_max=None,
                   ret_vals=True):
    if cheb:
        # in p low power first!!!
        p = np.polynomial.chebyshev.cheb2poly(p)

    # polyder have high power first
    dp = np.polyder(p[::-1])[::-1]

    x0 = np.polynomial.polynomial.polyroots(dp)
    x0 = x0[np.abs(np.imag(x0)) < 1e-4].real # Ok if we add unnecessary points

    x0 = [i for i in x0 if (i >= clip[0]) and i <= clip[1] ]

    if clip[0] > -np.inf and clip[0] not in x0:
        x0.append(clip[0])

    if clip[1] < +np.inf and clip[1] not in x0:
        x0.append(clip[1])

    vals = np.polynomial.polynomial.polyval(x0, p)
    if take_abs:
        vals = np.abs(vals)

    if k_max is None:
        idx = np.argmax(vals)
    else:
        idx = np.argsort(vals)[::-1][:k_max]

    x_ret = np.asarray(x0)[idx]

    if ret_vals:
        return x_ret, vals[idx]
    else:
        return x_ret


def _step_top_k(X_prev, G_prev, G, k, k_loc):
    # G_prev = num_points x rank
    num_gets = np.empty(G_prev.shape[0], dtype=int)

    all_x = []
    all_y = []

    for i, v in enumerate(G_prev):
        G_cur = np.einsum('i,ijk->jk', v, G)

        G0 = np.copy(G_cur)
        G0[0, :] /= 2**0.5

        p = sum([np.polynomial.chebyshev.Chebyshev(cf)**2 for cf in G0.T])
        p_poly = p.convert(kind=np.polynomial.Polynomial)

        x_max, y_max = _find_poly_max(list(p_poly), cheb=False,
            take_abs=True, k_max=k_loc, ret_vals=True)

        all_x.extend(x_max)
        all_y.extend(y_max)

        num_gets[i] = len(x_max)

    idx_maxx = np.argsort(all_y)[::-1][:k]
    k_real = len(idx_maxx)

    # trick as we flatten all y_max and (theoretically) we can get less points
    num_gets_cs = np.cumsum(num_gets)

    H_new_all = _cheb_my_poly(np.array(all_x)[idx_maxx], G.shape[1])

    G_new = np.empty([k_real, G.shape[2]])
    X_new = np.empty([k_real, 1 if X_prev is None else X_prev.shape[1] + 1 ])

    for idx, Hi, g_new, x_new in zip(idx_maxx, H_new_all, G_new, X_new):
        # idx_prev -- to what pull of prev points this max belongs
        idx_prev = np.searchsorted(num_gets_cs, idx, side='right')

        if X_prev is not None:
            x_new[:-1] = X_prev[idx_prev]

        x_new[-1] = all_x[idx]

        g_new[:] = np.einsum('i,j,ijk', G_prev[idx_prev], Hi, G)

    # ???  to_norm G_new???
    return X_new , G_new



"""Package teneva, module sample_func: sampling from functional TT-tensor.

This module contains the function "sample_func" for sampling from the
functional TT-tensor (i.e., the tensor of interpolation coefficients).

"""
import numpy as np
import teneva


def sample_func(A, seed=None, cores_are_prepared=False):
    """Sample random points according to given functional probability TT-tensor.

    Args:
        A (list): TT-tensor, which represents the interpolation coeeficients
            for the probability distribution.
        cores_are_prepared (bool): special flag for inner usage.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.

    Returns:
        np.ndarray: generated point for the tensor in the form of array of the
        shape [d], where d is the dimension.

    """
    rand = teneva._rand(seed)

    if not cores_are_prepared:
        A = teneva.copy(A)
        for G in A:
            G[:, 0, :] *= np.sqrt(2.)
        A = teneva.orthogonalize(A, 0)

    x_prev = None
    G_prev = np.array([1.])
    for G in A:
        G_cur = np.einsum('i,ijk->jk', G_prev, G)

        G0 = np.copy(G_cur)
        G0[0, :] /= np.sqrt(2)

        x_new = np.empty(1 if x_prev is None else x_prev.shape[0] + 1)
        if x_prev is not None:
            x_new[:-1] = x_prev

        p = sum([np.polynomial.chebyshev.Chebyshev(cf)**2 for cf in G0.T])
        x_cur = x_new[-1] = _sample_poly_1(p, rand)

        H_new = _cheb_my_poly(np.array(x_cur), G.shape[1])
        G_prev = np.einsum('i,ik', H_new, G_cur)
        x_prev = x_new

    return x_prev


def _cheb_my_poly(X, n):
    X = np.asarray(X)
    to_red = False
    if X.ndim == 0:
        X = X[None]
        to_red = True

    res = np.ones([X.shape[0], n])
    if n > 1:
        res[:, 1] = X

    for i in range(2, n):
        res[:, i] = 2*X*res[:, i - 1] - res[:, i - 2]

    res[:, 0] = np.sqrt(0.5)

    if to_red:
        res = res[0]

    return res


def _sample_poly_1(p2, rand):
    p2_int = p2.integ(lbnd=-1)
    xi = rand.uniform() # np.random.uniform() #
    p2_int_sh = p2_int - xi*(p2_int(1))
    roots = [np.real(i) for i in p2_int_sh.roots() if np.imag(i) == 0]
    assert len(roots) == 1
    return roots[0]



"""Package teneva, module anova_func: Functional ANOVA in the TT-format.

This module contains the function "anova_func" which computes the
TT-approximation of Chebyshev interpolation coefficients' tensor, using given
train dataset with random points.

"""
import numpy as np
import scipy as sp
import teneva


class ANOVA_func:
    def __init__(self, X_trn, y_trn, n, a=-1., b=+1., lamb=1.E-7):
        self.X_trn = teneva.poi_scale(X_trn, a, b, kind='cheb')
        self.y_trn = np.asarray(y_trn, dtype=float)
        self.lamb = lamb
        self._cfs = None
        self.d = self.X_trn.shape[1]
        self.n = n

    @property
    def coeffs(self):
        if self._cfs is not None:
            return self._cfs

        self._cfs = cfs = []

        y0 = np.mean(self.y_trn)
        cfs.append(y0)

        y = self.y_trn - y0
        for xd in self.X_trn.T:
            A = teneva.func_basis(xd, m=self.n, kind='cheb').T
            AtA = A.T @ A
            Aty = A.T @ y

            cur_cf = sp.linalg.lstsq(AtA + self.lamb * np.identity(A.shape[1]),
                Aty, overwrite_a=True, overwrite_b=True,
                lapack_driver='gelsy')[0]
            cfs.append(cur_cf[1:])

            cfs[0] += cur_cf[0]

        return self._cfs

    def cores(self, e=1.E-8):
        cfs = self.coeffs
        idx = np.zeros(self.d, dtype=int)
        A = teneva.delta([self.n]*self.d, idx, cfs[0])

        for i, cf in enumerate(cfs[1:]):
            idx[:] = 0
            for pi, p in enumerate(cf):
                idx[i] = pi + 1
                A = teneva.add(A, teneva.delta([self.n]*self.d, idx, p))

        return A if e is None else teneva.truncate(A, e)


def anova_func(X_trn, y_trn, n, a=-1., b=+1., lamb=1.E-7, e=1.E-8):
    """Build functional TT-tensor by TT-ANOVA from the given random points.

    Args:
        X_trn (np.ndarray): train points in the form of array of the shape
            [samples, d].
        y_trn (np.ndarray): values of the tensor for train points X_trn in the
            form of array of the shape [samples].
        n (int): mode size of the tensor (i.e, the maximum power of the
            interpolation coefficient plus 1).
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d or float).
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d or float).
        lamb (float): regularization parameter.
        e (float): optional truncation accuracy for resulting TT-tensor.

    Returns:
        list: TT-tensor, which represents the interpolation coefficients in the
        TT-format.

    Note:
        A class "ANOVA_func" that represents a wider set of methods for working
        with this decomposition is also available; see "anova_func.py".

    """
    return ANOVA_func(X_trn, y_trn, n, a, b, lamb).cores(e)



import numpy as np
import teneva
from time import perf_counter as tpc
import unittest


class TestMaxvolMaxvol(unittest.TestCase):
    def setUp(self):
        self.n = 5000   # Number of rows
        self.r = 50     # Number of columns
        self.e = 1.01   # Accuracy parameter
        self.k = 500    # Maximum number of iterations

        # Random tall matrix:
        rand = np.random.default_rng(42)
        self.A = rand.normal(size=(self.n, self.r))

        self.eps = 1.E-12

    def test_base(self):
        I, B = teneva.maxvol(self.A, self.e, self.k)
        C = self.A[I, :]

        e = np.max(np.abs(B))
        self.assertTrue(1. <= e <= 1.01)

        e = np.max(np.abs(self.A - B @ C))
        self.assertLess(e, self.eps)


class TestMaxvolMaxvolRect(unittest.TestCase):
    def setUp(self):
        self.n = 5000   # Number of rows
        self.r = 50     # Number of columns
        self.e = 1.01   # Accuracy parameter
        self.k = 500    # Maximum number of iterations

        # Random tall matrix:
        rand = np.random.default_rng(42)
        self.A = rand.normal(size=(self.n, self.r))

        self.eps = 1.E-12

    def test_base(self):
        dr_min = 2  # Minimum number of added rows
        dr_max = 8  # Maximum number of added rows

        I, B = teneva.maxvol_rect(self.A, self.e, dr_min, dr_max)
        C = self.A[I, :]

        e = np.max(np.abs(B))
        self.assertTrue(1. <= e <= 1.01)

        e = np.max(np.abs(self.A - B @ C))
        self.assertLess(e, self.eps)


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



"""Package teneva, module act_one: single TT-tensor operations.

This module contains the basic operations with one TT-tensor (Y), including
"copy", "get", "sum", etc.

"""
try:
    import numba as nb
    WITH_NUMBA = True
except Exception as e:
    WITH_NUMBA = False
import numpy as np
import teneva


def copy(Y):
    """Return a copy of the given TT-tensor.

    Args:
        Y (int, float, list): TT-tensor (or it may be int, float and numpy
            array for convenience).

    Returns:
        list: TT-tensor, which is a copy of the given TT-tensor. If Y is a
        number, then result will be the same number. If Y is np.ndarray, then
        the result will the corresponding copy in numpy format. If the
        function's argument is None, then it will also return None.

    """
    if Y is None or teneva._is_num(Y):
        return Y
    elif isinstance(Y, np.ndarray):
        return Y.copy()
    else:
        return [G.copy() for G in Y]


def get(Y, i, _to_item=True):
    """Compute the element (or elements) of the TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        i (list, np.ndarray): the multi-index for the tensor (list or 1D array
            of the length d) or a batch of multi-indices in the form of a list
            of lists or array of the shape [samples, d].

    Returns:
        float: the element of the TT-tensor. If argument i is a batch of
        multi-indices, then array of the length samples will be returned (the
        get_many function is called in this case).

    """
    d = len(Y)
    i = np.asanyarray(i, dtype=int)

    if i.ndim == 2:
        return get_many(Y, i)

    Q = Y[0][0, i[0], :]
    for k in range(1, d):
        Q = Q @ Y[k][:, i[k], :]

    return Q[0] if _to_item else Q


def get_and_grad(Y, i, check_phi=False):
    """Compute the element of the TT-tensor and gradients of its TT-cores.

    Args:
        Y (list): d-dimensional TT-tensor.
        i (list, np.ndarray): the multi-index for the tensor.
        check_phi (bool): service flag, should be False.

    Returns:
        (float, list): the element y of the TT-tensor at provided multi-index
        and the TT-tensor of related gradients for all TT-cores.

    """
    phi_r = interface(Y, i=i, norm=None, ltr=False)
    phi_l = interface(Y, i=i, norm=None, ltr=True)
    value = phi_r[0].item()

    if check_phi:
        # We check the correctness of the interfaces:
        p1 = phi_r[0].item()
        p2 = phi_l[-1].item()
        err = abs(val - p2)
        flag = (abs(p1) < 1e-8 and err < 1e-8) or err / abs(p1) < 1e-6
        text = f'Something unexpected, {p1}, {p2}, {err/abs(p1)}'
        assert flag, text

    grad = [np.zeros(G.shape) for G in Y]
    for Q, k, p_l, p_r in zip(grad, i, phi_l[:-1], phi_r[1:]):
        Q[:, k, :] = np.outer(p_l, p_r)

    return value, grad


def get_many(Y, I, _to_item=True):
    """Compute the elements of the TT-tensor on many indices (batch).

    Args:
        Y (list): d-dimensional TT-tensor.
        I (list of list, np.ndarray): the multi-indices for the tensor in the
            form of a list of lists or array of the shape [samples, d].

    Returns:
        np.ndarray: the elements of the TT-tensor for multi-indices I (array
        of the length samples).

    """
    d = len(Y)
    I = np.asanyarray(I, dtype=int)

    Q = Y[0][0, I[:, 0], :]
    for k in range(1, d):
        Q = np.einsum('iq, qir -> ir', Q, Y[k][:, I[:, k], :])

    return Q[:, 0] if _to_item else Q


def getter(Y, compile=True):
    """Build the fast getter function to compute the element of the TT-tensor.

    Args:
        Y (list): TT-tensor.
        compile (bool): flag, if True, then the getter will be called one time
            with a random multi-index to compile its code.

    Returns:
        function: the function that computes the element of the TT-tensor. It
        has one argument k (list or np.ndarray of the length d) which is
        the multi-index for the tensor.

    Note:
        Note that the gain from using this getter instead of the base function
        "get" appears only in the case of many requests for calculating the
        tensor value (otherwise, the time spent on compiling the getter may
        turn out to be significant). Also note that this function requires
        "numba" package to be installed.

        Attention: this function will be removed in the future! Use the
        "get_many" function instead (it's faster in most cases).

    """
    if not WITH_NUMBA:
        raise ValueError('Numba is required for this function')

    Y_nb = tuple([np.array(G, order='C') for G in Y])

    @nb.jit(nopython=True)
    def get(k):
        Q = Y_nb[0]
        y = [Q[0, k[0], r2] for r2 in range(Q.shape[2])]
        for i in range(1, len(Y_nb)):
            Q = Y_nb[i]
            R = np.zeros(Q.shape[2])
            for r1 in range(Q.shape[0]):
                for r2 in range(Q.shape[2]):
                    R[r2] += y[r1] * Q[r1, k[i], r2]
            y = list(R)
        return y[0]

    if compile:
        y = get(np.zeros(len(Y), dtype=int))

    return get


def interface(Y, P=None, i=None, norm='linalg', ltr=False):
    """Generate interface vectors for provided TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        P (list, np.ndarray): optional weights for mode indices from left to
            right (list of lists of the length d; or just one list if the
            weights are the same for all modes and all modes are equal).
        i (list, np.ndarray): optional multi-index for the tensor.
        norm (str): optional norm function to use (it may be 'linalg' ['l'] for
            the usage of the np.linalg.norm or 'natural' ['n'] for usage of the
            natural norm, i.e., the related mode size; or it may be None).
        ltr (bool): the direction of computation of the interface vectors
            ("left to right" if True and "right to left" if False).

    Returns:
        list: list of d+1 interface vectors. Note that the first and last
        vectors always have length 1.

    """
    d = len(Y)
    phi = [None] * (d+1)
    phi[-1] = np.ones(1)

    if ltr:
        Y = Y[::-1]
        if i is not None:
            i = i[::-1]
        if P is not None and not isinstance(P[0], (int, float)):
            P = P[::-1]

    for k in range(d-1, -1, -1):
        if i is None:
            if P is None:
                Q = np.sum(Y[k], axis=1)
            else:
                p = P if isinstance(P[0], (int, float)) else P[k]
                Q = np.einsum('rmq,m->rq', Y[k], p)
        else:
            if P is None:
                Q = Y[k][:, i[k], :]
            else:
                p = P if isinstance(P[0], (int, float)) else P[k]
                Q = Y[k][:, i[k], :] * p[i[k]]

        if ltr:
            Q = Q.T

        phi[k] = Q @ phi[k+1]

        if norm is not None:
            if norm.startswith('l'): # linalg
                phi[k] /= np.linalg.norm(phi[k])

            if norm.startswith('n'): # natural
                phi[k] /= Y[k].shape[1]

    if ltr:
        phi = phi[::-1]

    return phi


def mean(Y, P=None, norm=True):
    """Compute mean value of the TT-tensor with the given inputs probability.

    Args:
        Y (list): TT-tensor.
        P (list): optional probabilities for each dimension. It is the list of
            length d (number of tensor dimensions), where each element is
            also a list with length equals to the number of tensor elements
            along the related dimension. Hence, P[m][i] relates to the
            probability of the i-th input for the m-th mode (dimension).
        norm (bool): service (inner) flag, should be True.

    Returns:
        float: the mean value of the TT-tensor.

    """
    Z = np.ones((1, 1))
    for i in range(len(Y)):
        k = Y[i].shape[1]
        if P is None:
            p = np.ones(k) / k if norm else np.ones(k)
        else:
            p = P[i][:k]
        Z = Z @ np.einsum('rmq,m->rq', Y[i], p)
    return Z[0, 0]


def norm(Y, use_stab=False):
    """Compute Frobenius norm of the given TT-tensor.

    Args:
        Y (list): TT-tensor.
        use_stab (bool): if flag is set, then function will also return the
            second argument p, which is the factor of 2-power.

    Returns:
        float: Frobenius norm of the TT-tensor.

    """
    if use_stab:
        v, p = teneva.mul_scalar(Y, Y, use_stab=True)
        return np.sqrt(v) if v > 0 else 0., p/2
    else:
        v = teneva.mul_scalar(Y, Y)
        return np.sqrt(v) if v > 0 else 0.


def qtt_to_tt(Y, q):
    """Transform the QTT-tensor into a TT-tensor.

    Args:
        Y (list): QTT-tensor. It is d*q-dimensional tensor with mode size 2.
        q (int): quantization factor, i.e., the mode size of the TT-tensor will
            be n = 2^q.
    Returns:
        list: TT-tensor. It is d-dimensional tensor with mode size 2^q.

    """
    d = int(len(Y) / q)
    Z = []
    for k in range(d):
        G_list = Y[k*q:(k+1)*q]
        Z.append(teneva.core_qtt_to_tt(G_list))
    return Z


def sum(Y):
    """Compute sum of all tensor elements.

    Args:
        Y (list): TT-tensor.

    Returns:
        float: the sum of all tensor elements.

    """
    return mean(Y, norm=False)


def tt_to_qtt(Y, e=1.E-12, r=100):
    """Transform the TT-tensor into a QTT-tensor.

    Args:
        Y (list): TT-tensor. It is d-dimensional tensor with mode size n,
            which is a power of two, i.e., n=2^q.
        e (float): desired approximation accuracy.
        r (int): maximum rank for the SVD decomposition.

    Returns:
        list: QTT-tensor. It is d * q-dimensional tensor with mode size 2.

    """
    Z = []
    for G in Y:
        Z.extend(teneva.core_tt_to_qtt(G, e, r))
    return Z



import numpy as np
import teneva
from time import perf_counter as tpc
import unittest


class TestAlsAls(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [5] * self.d
        self.r = 3
        self.m = 1.E+4

        def func(I):
            return np.sum(I, axis=1)

        self.I_trn = teneva.sample_lhs(self.n, self.m, seed=42)
        self.y_trn = func(self.I_trn)

        self.I_tst = teneva.sample_rand(self.n, self.m, seed=42)
        self.y_tst = func(self.I_tst)

    def test_base(self):
        Y = teneva.rand(self.n, self.r, seed=42)
        Y = teneva.als(self.I_trn, self.y_trn, Y)

        e_trn = teneva.accuracy_on_data(Y, self.I_trn, self.y_trn)
        e_tst = teneva.accuracy_on_data(Y, self.I_tst, self.y_tst)

        self.assertLess(e_trn, 1.E-5)
        self.assertLess(e_tst, 1.E-5)

    def test_rank(self):
        Y = teneva.rand(self.n, 1, seed=42)
        Y = teneva.als(self.I_trn, self.y_trn, Y, r=10)

        e_trn = teneva.accuracy_on_data(Y, self.I_trn, self.y_trn)
        e_tst = teneva.accuracy_on_data(Y, self.I_tst, self.y_tst)

        self.assertLess(e_trn, 4.E-4)
        self.assertLess(e_tst, 4.E-4)

    def test_weight(self):
        Y = teneva.rand(self.n, self.r, seed=42)
        w = np.arange(self.m) + 1.
        Y = teneva.als(self.I_trn, self.y_trn, Y, w=w)

        e_trn = teneva.accuracy_on_data(Y, self.I_trn, self.y_trn)
        e_tst = teneva.accuracy_on_data(Y, self.I_tst, self.y_tst)

        self.assertLess(e_trn, 4.E-4)
        self.assertLess(e_tst, 4.E-4)


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



"""Package teneva, module act_many: operations with a set of TT-tensors.

This module contains the basic operations with a set of multiple TT-tensors
(Y1, Y2, ...), including "add_many" and "outer_many".

"""
import teneva


def add_many(Y_many, e=1.E-10, r=1.E+12, trunc_freq=15):
    """Compute Y1 + Y2 + ... + Ym in the TT-format.

    Args:
        Y_many (list): the list of TT-tensors (some of them may be int/float).
        e (float): desired approximation accuracy. The result will be truncated
            to this accuracy.
        r (int): maximum rank of the result.
        trunc_freq (int): frequency of intermediate summation result truncation.

    Returns:
        list: TT-tensor, which represents the element wise sum of all given
        tensors. If all the tensors are numbers, then result will be int/float.

    """
    Y = teneva.copy(Y_many[0])
    for i, Y_curr in enumerate(Y_many[1:]):
        Y = teneva.add(Y, Y_curr)
        if not teneva._is_num(Y) and (i+1) % trunc_freq == 0:
            Y = teneva.truncate(Y, e)
    return teneva.truncate(Y, e, r) if not teneva._is_num(Y) else Y


def outer_many(Y_many):
    """Compute outer product of given TT-tensors.

    Args:
        Y_many (list): list of TT-tensors.

    Returns:
        list: TT-tensor, which is the outer product of given TT-tensors.

    Note:
        See also "outer" function, which computes outer kronecker product of
        two given TT-tensors.

    """
    if len(Y_many) == 0:
        return None

    Y = teneva.copy(Y_many[0])
    for Y_curr in Y_many[1:]:
        Y.extend(teneva.copy(Y_curr))

    return Y



"""Package teneva, module core: operations with individual TT-cores.

This module contains functions to work directly with individual TT-cores.

"""
import numpy as np
import teneva


def core_dot(G, R, ltr=True):
    """Multiply TT-core G with matrix R."""
    # TODO: Add docs and demo.
    r1, n, r2 = G.shape
    G = teneva._reshape(G, (r1*n, r2) if ltr else (r1, n*r2))
    R = np.array([[R]]) if isinstance(R, (int, float)) else R
    G = G @ R if ltr else R @ G
    G = teneva._reshape(G, (r1, n, G.shape[1]) if ltr else (G.shape[0], n, r2))
    return G


def core_dot_inv(G, R, ltr=True):
    """Multiply TT-core G with inverted matrix R."""
    # TODO: Add docs and demo.
    r1, n, r2 = G.shape
    G = teneva._reshape(G, (r1*n, r2) if ltr else (r1, n*r2))
    G = np.linalg.solve(R.T, G.T).T if ltr else np.linalg.solve(R, G)
    return teneva._reshape(G, (r1, n, r2))


def core_dot_maxvol(G, R, ind=None, ltr=True):
    """Multiply TT-core G with matrix R and leaves the most important rows."""
    # TODO: Add docs and demo.
    r1, n, r2 = G.shape
    G = core_dot(G, R, ltr)
    G = teneva._reshape(G, (r1, n*G.shape[-1]) if ltr else (G.shape[0]*n, r2))
    ind = teneva._maxvol(G.T if ltr else G)[0] if ind is None else ind
    G = G[:, ind] if ltr else G[ind, :]
    return G, ind


def core_qr_rand(G, m, ltr=True, seed=None):
    """Add random rows to TT-core G and return the new TT-core from Q-factor."""
    # TODO: Add docs and demo.
    r1, n, r2 = G.shape

    rand = teneva._rand(seed)
    rnd = rand.normal(size=(r1*n if ltr else n*r2, m))

    G = teneva._reshape(G, (r1*n, r2) if ltr else (r1, n*r2))
    G = G if ltr else G.T
    G = np.hstack((G, rnd))
    G, _ = np.linalg.qr(G)
    G = G if ltr else G.T
    G = teneva._reshape(G, (r1, n, G.shape[1]) if ltr else (G.shape[0], n, r2))
    return G


def core_qtt_to_tt(Q_list):
    """Transform the list of QTT-cores into a TT-core.

    Args:
        Q_list (list of  np.ndarray): list of QTT-cores of the shapes [[q_0,
            2, q_1], [q_1, 2, q_2], ...[q_(q-1), 2, q_q]] and length q.

    Returns:
        np.ndarray: TT-core in the form of 3-dimensional array of the shape
        q_0 x 2^q x q_q.

    """
    G = Q_list[0].copy()

    for Q in Q_list[1:]:
        r1 = G.shape[0]
        r2 = Q.shape[-1]
        G = np.tensordot(G, Q, 1)
        G = teneva._reshape(G, (r1, -1, r2))

    return G


def core_stab(G, p0=0, thr=1.E-100):
    """Scaling for the passed TT-core, i.e., G -> (Q, p), G = 2^p * Q.

    Args:
        G (np.ndarray): TT-core in the form of 3-dimensional array.
        p0 (int): optional initial value of the power-factor (it will be added
            to returned value p).
        thr (float): threshold value for applying scaling (if the maximum
            modulo element in the TT-core is less than this value, then scaling
            will not be performed).

    Returns:
        (np.ndarray, int): scaled TT-core (Q) and power-factor (p), such
        that G = 2^p * Q.

    """
    v_max = np.max(np.abs(G))

    if v_max <= thr:
        return G, p0

    p = int(np.floor(np.log2(v_max)))
    Q = G / 2.**p

    return Q, p0 + p


def core_tt_to_qtt(G, e=0., r=1.E+12):
    """Transform the TT-core into a list of QTT-cores.

    Args:
        G (np.ndarray): TT-core in the form of 3-dimensional array of the shape
            r1xnxr2. The mode size should be a power of two, i.e., n=2^d.
        e (float): desired approximation accuracy.
        r (int): maximum rank for the SVD decomposition.

    Returns:
        list: list of QTT-cores (np.ndarrays) of the shape [q1, 2, q2], which
        approximates the given TT-core G.

    """
    r1, n, r2 = G.shape
    d = int(np.log2(n))
    if 2**d != n:
        raise ValueError('Invalid mode size (it should be a power of two)')

    A = teneva._reshape(G, (-1, r2))
    A, V0 = teneva.matrix_svd(A, e, r)

    Y = []
    for i in range(d-1):
        As = A.shape[0] // 2
        q = A.shape[1]
        A = np.hstack([A[:As], A[As:]])
        A, V = teneva.matrix_svd(A, e, r)
        Y.append(teneva._reshape(V, (-1, 2, q), order='C'))

    Y.append(teneva._reshape(A, (r1, 2, -1)))
    Y[0] = np.einsum('ijk,kl', Y[0], V0)

    return Y[::-1]
